import os
import warnings
import numpy as np
import multiprocessing as mp
from Bio import AlignIO, SeqIO, BiopythonWarning, codonalign
from collections import defaultdict
from mycotools.gff2seq import ntmain
from mycotools.lib.kontools import multisub
from mycotools.lib.biotools import gff2list, fa2dict, dict2fa
from orthocluster.orthocluster.lib import input_parsing
warnings.simplefilter('ignore', BiopythonWarning)

def dndsGeneGrab(
    gff_path, assembly_path, proteome_path,
    ome_sig_clus, gene2hg, clusplusminus
    ):
    
    gff_list, prot_dict = gff2list(gff_path), fa2dict(proteome_path)
    assem_dict, clus_out = fa2dict(assembly_path), {}
    cds_dict, cds_dict2 = input_parsing.compileCDS2(
        gff_list, os.path.basename(gff_path).replace('.gff3','')
        )
    
    for scaf in cds_dict:
        for i0, seq0 in enumerate(cds_dict[scaf]):
            if seq0 in ome_sig_clus: # if the og is part of a significant seed
            # locus
                try:
                    og0 = gene2hg[seq0]
                except KeyError:
                    continue
                for hit in ome_sig_clus[seq0]:
                    hgx = tuple(sorted(list(hit)))
                    if hgx not in clus_out:
                        clus_out[hgx] = [{og: [{}, {}] for og in hgx}]
                    clus_out[hgx][-1][og0][0][seq0] = prot_dict[seq0]
                    clus_out[hgx][-1][og0][1] = {
                        **clus_out[hgx][-1][og0][1], 
                        **ntmain(cds_dict2[seq0], assem_dict)
                        }
                    
                    start, end = None, None
                    if i0 < clusplusminus: # is i0 - clusplusminus < 0 ?
                        locus = cds_dict[scaf][:i0+clusplusminus+1] # then gather
                        # all the beginning
                    else:
                        locus = cds_dict[scaf][i0-clusplusminus:i0+clusplusminus+1]
                        # instead get the +/- and adjust for python
                    for i1, seq1 in enumerate(locus): # for each index and sequence
                    # in the locus
                        try:
                            og = gene2hg[seq1]
                        except KeyError:
                            continue
                        if og in hit: # if the og is
                        # in the sig clus
                            clus_out[hgx][-1][og][0][seq1] = prot_dict[seq1]
                            clus_out[hgx][-1][og][1] = {
                                **clus_out[hgx][-1][og][1],
                                **ntmain(cds_dict2[seq1], assem_dict)
                                }
    return clus_out



def dnds_preparation(db, omes4hgs, hgx2omes, gene2hg, plusminus, hgx_dir, i2ome):
    
    gene_checks, hgx_check = {}, {}
    
    for ome in omes4hgs:
        gff_path = db[i2ome[ome]]['gff3'] 
        proteome_path = db[i2ome[ome]]['faa']
        assembly_path = db[i2ome[ome]]['fna']
        res = dndsGeneGrab(
            gff_path, assembly_path, proteome_path,
            omes4hgs[ome], gene2hg, plusminus
            )
        for hgx in res:
            if hgx not in hgx_check:
                hgx_check[hgx] = set()
            if hgx not in gene_checks:
                gene_checks[hgx] = {og: [{}, {}] for og in hgx}
            for loc in res[hgx]:
                for og in loc:
                    gene_checks[hgx][og][0] = {
                        **gene_checks[hgx][og][0], **loc[og][0]
                        }
                    gene_checks[hgx][og][1] = {
                        **gene_checks[hgx][og][1], **loc[og][1]
                        }
            hgx_check[hgx].add(ome)
            if hgx_check[hgx] == set(hgx2omes[hgx]):
                for og in hgx: 
                    out_base = hgx_dir + '-'.join([str(x) for x in hgx]) + \
                        '.' + str(og)
                    with open(out_base + '.aa.fa', 'w') as out:
                        out.write(dict2fa(gene_checks[hgx][og][0]))
                    with open(out_base + '.nt.fa', 'w') as out:
                        out.write(dict2fa(gene_checks[hgx][og][1]))
                del gene_checks[hgx]
    
    if gene_checks:
        print('\t\tERROR: discrepancies with previous run', flush = True)
        for hgx in gene_checks:
            print('\t\t\t' + str(hgx), flush = True)
            for og in hgx: 
                out_base = hgx_dir + '-'.join([str(x) for x in hgx]) + \
                    '.' + str(og)
                with open(out_base + '.aa.fa', 'w') as out:
                    out.write(gene_checks[hgx][og][0]) 
                with open(out_base + '.nt.fa', 'w') as out:
                    out.write(gene_checks[hgx][og][1])



def calc_dnds(mafft):
    
    og_catch = re.search(r'(^[^\.]+)\.(\d+)\.mafft$', os.path.basename(mafft))
    hgx = tuple([int(x) for x in og_catch[1].split('-')])
    og = int(og_catch[2])
    
    err = os.path.basename(mafft).replace('.mafft','')
    nt_path = mafft.replace('.mafft', '.nt.fa')
    try:
        prot_align = AlignIO.read(mafft, 'fasta')
    except ValueError: 
        print('\t\t' + err + ' empty alignment?', flush = True)
        return None
    
    nt_fa = SeqIO.parse(nt_path, 'fasta')
    
    try:
        aln = codonalign.build(prot_align, nt_fa) #, ids)
    except KeyError: 
        print('\t\t' + err + ' ambiguous AAs', flush = True)
        return
    except RuntimeError:
        print('\t\t' + err + ' BioPython error', flush = True)
        return
    except IndexError: 
        print('\t\t' + err + ' unknown error', flush = True)
        return
    try:
        dn_matr, ds_matr = aln.get_dn_ds_matrix()
        np_dn, np_ds = np.array(dn_matr), np.array(ds_matr)
        dnds_matrNulls = np.divide(np_dn, np_ds)
        dnds_matr = np.ma.array(dnds_matrNulls, mask=np.isnan(dnds_matrNulls))
        dnds_matr = np.abs(dnds_matr)
        dnds = np.mean(dnds_matr)
        return hgx, og, dnds
    except ZeroDivisionError:
        print('\t\t' + err + ' unknown error')
    except KeyError: 
        print('\t\t' + err + ' ambiguous NTs')
    return None


def parse_dnds(mpRes):

    hgxdNdS_dict = {}
    for res in mpRes:
        if res:
            hgx, og, dnds = res[0], res[1], res[2]
            try:
                float(dnds)
            except ValueError:
                continue
            if hgx not in hgxdNdS_dict:
                hgxdNdS_dict[hgx] = {
                    'selection_total': 0,
                    'og_total': 0,
                    'hgs': {}
                    }

            if dnds < 1: # if purifying selection
                hgxdNdS_dict[hgx]['selection_total'] += dnds
            else: # if positive selection use reciprocal
                hgxdNdS_dict[hgx]['selection_total'] += 1/dnds
            hgxdNdS_dict[hgx]['og_total'] += 1
            hgxdNdS_dict[hgx]['hgs'][og] = dnds
    for hgx in hgxdNdS_dict:
        selection_coefficient = hgxdNdS_dict[hgx]['selection_total']/hgxdNdS_dict[hgx]['og_total']
        mean_dnds = sum(hgxdNdS_dict[hgx]['hgs'].values())/len(hgxdNdS_dict[hgx]['hgs'])
        hgxdNdS_dict[hgx] = [selection_coefficient, mean_dnds, hgxdNdS_dict[hgx]['hgs']]

    return hgxdNdS_dict

def dnds_main(db, ome2i, gcfs, hgx_dir, i2ome, hx2omes, gene2hg, plusminus):
    print('\nIX. Quantifying GCF dn/ds', flush = True)
    omes4hgs, hgx_files = {x: defaultdict(list) for x in range(len(ome2i))}, []
    for i, gcf in enumerate(gcfs):
        for hgx in gcf:
            hgx = hgx2i[x]
            for gene in hgx2loc[hgx]:
                omeI = ome2i[gene[:gene.find('_')]]
                if omeI in set(gcf[i][hgx]):
                    omes4hgs[omeI][gene].append(gcf_hgxs[i])

    hgx_files = []
    for hgx in gcf_hgxs:
        for og in hgx:
            out_base = hgx_dir + '-'.join([str(x) for x in hgx]) + '.' + str(og)
            hgx_files.append(out_base)

    maffts = [file + '.mafft' for file in hgx_files]
    if not all(os.path.isfile(x + '.aa.fa') for x in hgx_files):
        print("\tCompiling HGx kernels' genes", flush = True)
        dnds_preparation(db, omes4hgs, hgx2omes, gene2hg, plusminus, hgx_dir, i2ome)
    if not all(os.path.isfile(x) for x in maffts):
        print('\tAligning proteins', flush = True)
        maffts = [x for x in maffts if not os.path.isfile(x)]
        mafft_cmds = [
            [['mafft', '--auto', '--thread', '2', x.replace('.mafft', '.aa.fa')], x,
            x.replace('.mafft', '.aa.fa')] \
            for x in maffts
            ]
        mafft_res = multisub(mafft_cmds, processes = cpus, stdout = True, rm = True)
        maffts = [x['output'] for x in mafft_res]

    print('\tCalculating dn/ds', flush = True)
    with mp.get_context('fork').Pool(processes = cpus) as pool:
        dnds_res = pool.map(calc_dnds, maffts)
        pool.close()
        pool.join()

    hgx2dnds = parse_dnds(dnds_res)



#! /usr/bin/env python3

import os
import re
import sys
import argparse
import numpy as np
import multiprocessing as mp
from math import log
from graph_tool import centrality, clustering
from graph_tool.all import *
from scipy.sparse import lil_matrix
from itertools import combinations, chain
from collections import Counter, defaultdict
from cloci.lib.hgx2hlgs import MCL, read_clus, read_mci_rows
from mycotools.lib.kontools import format_path, getColors, hex2rgb, \
    write_json, read_json, eprint, mkOutput
from mycotools.lib.dbtools import mtdb
from mycotools.lib.biotools import gff2list
from mycotools.utils.extractHmmAcc import grabAccs

# NEED to output json of the network modules
# NEED to organize output
def collectGffs(omes, top_dir):
    clusGFFs = [
        top_dir + ome + '/' + ome + '.gff3' for ome in omes \
        if os.path.isfile(top_dir + ome + '/' + ome + '.gff3')
        ]
    return clusGFFs

def mkPfam2name(pfam_file):
    # compile pfam 2 name
    with open(pfam_file, 'r') as raw:
        pfam_str = raw.read()
    pfams = grabAccs(pfam_str)
    pfam2name = {x[1].lstrip(): x[0].lstrip() for x in pfams}
    pfam2name = {k[:k.find('.')]: v for k,v in pfam2name.items()}
    return pfam2name


def extract_lineages(db, rank):
    # acquire all lineages that belong to rank within input database
    omesXlineage = {}
    for ome, row in db.set_index().items():
        lineage = row['taxonomy'][rank]
        if lineage not in omesXlineage:
            omesXlineage[lineage] = []
        omesXlineage[lineage].append(ome)
    return omesXlineage


def parse_info(info_path, ome):
    # acquire all homology_group pairs from each cluster, return ready for counter hash

    counts, hg_counts, hg2pfam = Counter(), Counter(), defaultdict(list)
    with open(info_path, 'r') as raw:
        for line in raw:
            todel = []
            if not line.startswith('#'):
                data = line.rstrip().split('\t')
                hgs, genes = data[1].split(','), data[2].split(',')
                if not len(data) == 5:
                    eprint(f'\tWARNING: skipping {ome} missing annotations', flush = True)
                    return ome, None, None, None
                pfams = [x.split('|') for x in data[-1].split(';')]
                gene2pfam = {k: pfams[i] for i, k in enumerate(genes)}
                for i, hg in enumerate(hgs):
                    try:
                        hgs[i] = int(hg)
                    except ValueError:
                        todel.append(i)
                if todel:
                    for i in reversed(todel):
                        del hgs[i]
                        del pfams[i]

                for i, hg in enumerate(hgs):
                    if pfams[i]:
                        hg2pfam[hg].extend(list(set(pfams[i])))
                sort_hgs = sorted(hgs)
                hg_counts += Counter(set(sort_hgs))
                pairs = combinations(sort_hgs, 2)
                counts += Counter(pairs)
                

    return ome, tuple([tuple([k, v]) for k,v in counts.items()]), \
        tuple([tuple([k,v]) for k,v in hg_counts.items()]), \
        tuple([tuple([k,tuple(v)]) for k,v in hg2pfam.items()])


def combine_counts(counts_res, db, rank = 'species'):
    # combine counts from multiprocessing
    db = db.set_index()

    counts = defaultdict(Counter)
    hg_counts = defaultdict(Counter)
    hg2pfam = defaultdict(dict)
    taxonLens = defaultdict(int)
    taxon_fails = defaultdict(int)
    for ome, tCounts, tHGcounts, tHG2Pfam in counts_res:
        taxon = db[ome]['taxonomy'][rank]
        if tCounts == None:
            taxon_fails[taxon] += 1
            continue
        taxonLens[taxon] += 1
        counts[taxon] += Counter({x[0]: x[1] for x in tCounts})
        hg_counts[taxon] += Counter({x[0]: x[1] for x in tHGcounts})
        for hg, pfams in tHG2Pfam:
            if hg not in hg2pfam[taxon]:
                hg2pfam[taxon][hg] = Counter()
            hg2pfam[taxon][hg] += Counter(pfams)

    fCounts, fHGcounts, fHG2Pfam = Counter(), Counter(), {}
    for taxon in counts:
        # average across the rank
        tCounts = {k: round(v/taxonLens[taxon] + 0.5) for k, v in counts[taxon].items()}
        tHGcounts = {k: round(v/taxonLens[taxon] + 0.5) for k, v in hg_counts[taxon].items()}
        tHG2Pfam = {}
        for hg, pfams in hg2pfam[taxon].items():
            tHG2Pfam[hg] = {k: round(v/taxonLens[taxon]) for k, v in pfams.items()}

        fCounts += Counter(tCounts)
        fHGcounts += Counter(tHGcounts)
        for hg, pfams in tHG2Pfam.items():
            if hg not in fHG2Pfam:
                fHG2Pfam[hg] = Counter()
            fHG2Pfam[hg] += pfams

    for hg in fHG2Pfam:
        fHG2Pfam[hg] = {k: v for k,v in sorted(fHG2Pfam[hg].items(), key = lambda x: x[1], reverse = True)}
            
    return fCounts, fHGcounts, fHG2Pfam, taxon_fails

def write_node_quants(hg_counts, out_path):

    hg_counts = [
        (k, v) for k,v in sorted(hg_counts.items(), key = lambda x: x[1],
        reverse = True)
        ]
    with open(out_path, 'w') as out:
        for k,v in hg_counts:
            out.write(str(k) + '\t' + str(v) + '\n')


def increase_float(val):
    # high fidelity float increase
    strVal = str(val)
    if strVal.endswith('9'):
        newVal = float(strVal + '1')
    else:
        factor = len(strVal[strVal.find('.')+1:])
        scale = 10**factor
        transVal = round(val*scale)
        newVal = float('0.' + str(transVal + 1))
    return newVal


def write_to_adj(oEdges, edges, out_path, oi2ni):
    minCount, maxCount = min(oEdges.values()), max(oEdges.values())
    denom = maxCount - minCount
    if denom == 0:
        eprint('\t\t\tWARNING: low connection weight throughout network', flush = True)
        denom = 1
    with open(out_path, 'w') as outF:
        for k, value in edges.items():
#            value = ((v-minCount)/denom) # normalize
            out_data = [oi2ni[k[0]], oi2ni[k[1]], value]
            outF.write('\t'.join([str(x) for x in out_data]) + '\n')



def write_adj_matr(oEdges, out_path, maximum = 500, duplicates = False):

    if not duplicates: # dont look at same HG-HG relationships
        oEdges = {k: v for k,v in oEdges.items() if k[0] != k[1]}

    data = sorted(list(oEdges.values()))
    edges = oEdges
    
#    write_to_adj(oEdges, edges, out_path + '.raw', oi2ni)

    percentile = 0.0
    while len(edges) > maximum: # while too many edges
        percentile = increase_float(percentile)
        minPerc = data[round((len(data)-1)*percentile)]
        edges = {k: v for k, v in oEdges.items() if v > minPerc}

    edge_ns = sorted(set(chain(*list(oEdges.keys()))))
    ni2oi = {int(k): int(v) for k, v in enumerate(edge_ns)}
    oi2ni = {v: k for k, v in ni2oi.items()}

    print('\t\tEdge #|percentile:', len(edges), percentile, flush = True)
    write_to_adj(oEdges, edges, out_path, oi2ni)
    
    return ni2oi, oi2ni    


def import_adj(adj_file, loclen, minimum = 0):
#    adj_arr = np.zeros([len(loci2grab), len(loci2grab)])
    adj_arr = lil_matrix((loclen, loclen))

    count = 0
    with open(adj_file, 'r') as raw:
        for line in raw:
            l0, l1, gid = line.rstrip().split()
            l0, l1 = int(l0), int(l1)
            val = float(gid)
            if val >= minimum:
                adj_arr[(l0, l1)] = val
                
    return adj_arr


def run_community_detection(wrk_dir, adj_f, inflation = 1.5, threads = 1):
    clus_f = adj_f + '.clus'
    rows_f = adj_f + '.rows'
    MCL(adj_f, clus_f, inflation, threads, rows_f)
    mci2pre = read_mci_rows(rows_f)
    clus = read_clus(clus_f, mci2pre)
    i2clus = {}
    for i,v in enumerate(clus):
        for i1 in v:
            i2clus[i1] = i
    return clus, i2clus


def make_network(
    wrk_dir, adj_file, quant_path, network_path, hg2pfam_path,
    ni2oi, oi2ni, calc_transitivity = False, calc_centrality = False,
    calc_local_transitivity = False,
    skip = False, overwrite = False, scale = 1.5, def_size = 1, annotate = False,
    inflation = 1.5, cpus = 1
    ):
    # open adjacency matrix and construct a network as an html


    clus, i2clus = run_community_detection(wrk_dir, adj_file, 
                                        inflation, cpus * 2 - 1)

    adj_arr = import_adj(adj_file, len(oi2ni))
    idx = adj_arr.nonzero()
    g = Graph(directed = False)
    g.add_edge_list(np.transpose(idx))
    weights = adj_arr[idx].toarray()
    reprop = g.new_edge_property('float')
    reprop.a = weights
    v = weights[:, 1]
    weights[:, 1] = log((v - v.min()) / (v.max() - v.min()) + def_size)
    eprop = g.new_edge_property('float')
    eprop.a = weights

    quants_p = {}
    with open(quant_path, 'r') as raw:
        for line in raw:
            k, v = line.rstrip().split('\t')
            if int(k) in oi2ni: # HGs with only duplicates may be excluded
                quants_p[oi2ni[int(k)]] = float(v)


    max_q, min_q = max(quants_p.values()), min(quants_p.values())
    denom = max_q - min_q
    quants = g.new_vertex_property('float')
    if denom:
        for k, v in quants_p.items():
            quants[k] = log((v - min_q)/denom + def_size) + def_size * 3
    else:
        for k, v in quants_p.items():
            quants[k] = 4

    hg2pfam_dict = {}
    hg2pfam = g.new_vertex_property('string')
    if os.path.isfile(hg2pfam_path):
        hg2pfamPrep = read_json(hg2pfam_path)
        for hg, pfams in hg2pfamPrep.items():
            hg = int(hg)
            if hg in oi2ni:
                if pfams:
                    v = list(pfams.keys())[0]
                    hg2pfam_dict[oi2ni[hg]] = f'{hg}|{v[v.find("-")+1:]}'
                    hg2pfam[oi2ni[hg]] = f'{hg}|{v[v.find("-")+1:]}'
                else:
                    hg2pfam[oi2ni[hg]] = str(hg)
                    hg2pfam_dict[oi2ni[hg]] = str(hg)
            
    else:
        for k, v in enumerate(quants):
            hg2pfam[k] = str(ni2oi[k])
            hg2pfam_dict[k] = str(ni2oi[k])


#    if len(quants) < 2:
 #       eprint('\t\tERROR: less than 2 nodes in network', flush = True)
  #      return
    count = 0
    v_fil_col = g.new_vertex_property('vector<float>')
    colors = getColors(len(clus), ignore = ['#ffffff'])
    for c, nodes in enumerate(clus):
        if len(nodes) > 1:
            try:
                color = colors[count]
                count += 1
            except IndexError:
                count = 0
                color = colors[count]
        else:
            color = '#ffffff'
        rgb = [x/255.0 for x in hex2rgb(color)] + [0.8]
        for hg in nodes:
#            if hg in oi2ni:
 #               ni = oi2ni[hg]
            v_fil_col[hg] = rgb
        

    g.vp['name'] = hg2pfam
    print('\t\tWriting', flush = True)
    if not skip:
        if overwrite or not os.path.isfile(network_path):
            if annotate:
                graph_draw(g, vertex_size = quants, vertex_text = g.vp['name'],
                   output = network_path, vertex_halo_color = v_fil_col, 
                   vertex_fill_color = v_fil_col,
                   edge_pen_width = eprop, edge_color = [0,0,0,1], vertex_font_family = 'arial',
                   vertex_font_size = 1)
            else:
                graph_draw(g, vertex_size = quants,
                   output = network_path, vertex_halo_color = v_fil_col, 
                   vertex_fill_color = v_fil_col,
                   edge_pen_width = eprop, edge_color = [0,0,0,1])
    
    print('\t\tNodes:', len(oi2ni), flush = True)
#    print('\t\tEdges:', len(edgeLabels), flush = True)


    print('\t\tQuantifying graph summary', flush = True)
    if calc_transitivity:
        print('\t\t\tQuantifying global graph transitivity', flush = True)
        print('\t\t\t\tNode:', flush = True)
        global_transitivity = clustering.global_clustering(g, reprop)
        stats = global_transitivity
        print(f'\t\t\t\t\t{global_transitivity[0]} stdev: {global_transitivity[1]}', 
              flush = True)
    else:
        stats = [None, None]

    density = len(weights[0])/((len(quants_p) - (len(quants_p) - 1))/2)
    stats = stats + (density,)
    print(f'\t\t\tDensity: {density}', flush = True)

    if calc_local_transitivity:
        local_transitivity = clustering.local_clustering(g, reprop)
        trans_path = adj_file[:-4] + '.transitivity'
        with open(trans_path, 'w') as out:
            out.write('#hg|pfam\ttransitivity\tnode_size\n')
            for k, v in sorted(enumerate(local_transitivity),
                               key = lambda x: x[1], reverse = True):
                if k in quants_p:
                    out.write(f'{hg2pfam_dict[k]}\t{v}\t{quants_p[k]}\n')
                elif k in hg2pfam_dict:
                    eprint(f'\t\t\tWARNING: {hg2pfam_dict[k]} missing from node sizes',
                           flush = True)
                else:
                    eprint(f'\t\t\tWARNING: rogue network node {k}', flush = True)
#        print('\t\t\t\tEdge:', flush = True)
 #       edge_transitivity = nx.edge_transitivity(G)
  #      print('\t\t\t\t\t' + str(edge_transitivity), flush = True)
    if calc_centrality:
        print('\t\t\tQuantifying betweenness centrality', flush = True)
        v_between, e_between = centrality.betweenness(g)
        central_path = adj_file[:-4] + '.centrality'
        with open(central_path, 'w') as out:
            out.write('#hg|pfam\tbetweenness\tnode_size\n')
            for k, v in sorted(enumerate(v_between),
                               key = lambda x: x[1], reverse = True):
                if v == 0:
                    continue
                if k in quants_p:
                    out.write(f'{hg2pfam_dict[k]}\t{v}\t{quants_p[k]}\n')
                elif k in hg2pfam_dict:
                    eprint(f'\t\t\tWARNING: {hg2pfam_dict[k]} missing from node sizes',
                           flush = True)
                else:
                    eprint(f'\t\t\tWARNING: rogue network node {k}', flush = True)

    return stats 

def main(
    db, rank, cloci_dir, transitivity = False, 
    calc_centrality = False, local_transitivity = False,
    max_edges = 500, hlg = 'hlg',
    overwrite = False, cpus = 1, skip = False
    ):

    if cloci_dir:
        ome_dir, net_dir = cloci_dir + 'ome/', cloci_dir + 'net/'
        if not os.path.isdir(ome_dir):
            eprint('\nERROR: invalid cloci output', flush = True)
            sys.exit(2)

    wrk_dir = net_dir + 'working/'
    if not os.path.isdir(net_dir):
        os.mkdir(net_dir)
    if not os.path.isdir(wrk_dir):
        os.mkdir(wrk_dir)
    
    omesXlineage = extract_lineages(db, rank)

    omesXlineage = {
        k: v for k,v in sorted(omesXlineage.items(), key = lambda x: len(x[1]))
        }
    lineage2stats = {}
    for lineage, omes in omesXlineage.items():
        print('\n' + lineage, flush = True)
        adj_path = wrk_dir + lineage + '.adj'
        quant_path = wrk_dir + lineage + '.tsv'
        hg2pfam_path = wrk_dir + lineage + '.hg2pfam.txt'
#        if not os.path.isfile(adj_path): # need to log the factor then
        print('\tAdjacency matrix', flush = True)
        info_cmds = [
            [f'{ome_dir}{ome}/{hlg}.tsv', ome] \
            for ome in omes \
            if os.path.isfile(f'{ome_dir}{ome}/{hlg}.tsv')
            ]
        if not info_cmds:
            eprint('\tWARNING: skipping; no output files found', flush = True)
            continue
        with mp.Pool(processes = cpus) as pool:
            counts_res = pool.starmap(parse_info, info_cmds)
        counts, hg_counts, hg2pfam, taxon_fails = combine_counts(
            counts_res, db, rank = 'species'
            )
        if not hg_counts:
            eprint('\tWARNING: skipping; no annotations', flush = True)
            continue
        write_json(hg2pfam, hg2pfam_path)
        write_node_quants(hg_counts, quant_path)
        ni2oi, oi2ni = write_adj_matr(counts, adj_path, max_edges)
    
        net_path = net_dir + lineage + '.svg'
        print('\tNetwork', flush = True)
        stats = make_network(wrk_dir, adj_path, quant_path, net_path, 
                         hg2pfam_path, ni2oi, oi2ni,
                         transitivity, calc_centrality, 
                         local_transitivity, skip,
                         overwrite)
        lineage2stats[lineage] = stats + (len(omes) - taxon_fails[lineage], taxon_fails[lineage],)

    with open(net_dir + rank + '.stats.tsv', 'w') as out:
        out.write('#lineage global_transitivity gl_tr_stdev density pass_genomes failed_genomes\n')
        for lineage, stats in lineage2stats.items():
            out.write(f'{lineage} {" ".join([str(x) for x in stats])}\n')
     
            

def cli():
    parser = argparse.ArgumentParser(
        description = 'Takes CLOCI output, creates networks of HG' + \
        ' relationships. Node sizes are normalized and edge sizes are ' + \
        'log-normalized during visualization'
        )
    parser.add_argument(
        '-i', '--input', required = True,
        help = 'CLOCI results directory'
        )
    parser.add_argument(
        '-g', '--gcfs', help = 'Make networks from GCFs; DEFAULT HLGs',
        action = 'store_true',
        )
    parser.add_argument(
        '-d', '--database', required = True, 
        help = 'Must be same as CLOCI input'
        )
    parser.add_argument(
        '-r', '--rank', help = 'Taxon rank; DEFAULT: class',
        default = 'class'
        )
    parser.add_argument('-s', '--skip', help = 'Skip writing network file', 
                        action = 'store_true')
    parser.add_argument(
        '-m', '--maximum', help = 'Find upper percentile of edges to fit ' + \
        'maximum edges',
        type = int
        )
    parser.add_argument(
        '--centrality', help = 'Calculate betweenness centrality',
        action = 'store_true'
        )
    parser.add_argument(
        '-gt', '--global_transitivity', help = 'Quantify global transitivity',
        action = 'store_true'
        )
    parser.add_argument('-lt', '--local_transitivity', help = 'Quantify local transitivity',
                        action = 'store_true')
    parser.add_argument(
        '-c', '--cpu', default = 1, type = int
        )
    parser.add_argument('--overwrite', help = 'Rerun if necessary',
        action = 'store_true')
    args = parser.parse_args()

    db = mtdb(format_path(args.database))
    cloci_dir = format_path(args.input)

    if args.gcfs:
        hlg = 'gcf'
    else:
        hlg = 'hlg'

    main(
        db, args.rank, cloci_dir, transitivity = args.global_transitivity, 
        calc_centrality = args.centrality, local_transitivity = args.local_transitivity,
        hlg = hlg,
        max_edges = args.maximum, cpus = args.cpu, overwrite = args.overwrite, 
        skip = args.skip
        )
    sys.exit(0)


if __name__ == '__main__':
    cli()



#! /usr/bin/env python3

import os
import re
import sys
import gzip
import argparse
import multiprocessing as mp
from math import log
from tqdm import tqdm
from collections import defaultdict, Counter
from cloci.lib.output_data import run_hmmsearch, parse_hmm_res
from mycotools.annotationStats import main as annotation_stats
from mycotools.assemblyStats import main as assembly_stats
from mycotools.lib.kontools import collect_files, eprint, format_path, findExecs, mkOutput
from mycotools.lib.biotools import gff3Comps, gff2list
from mycotools.lib.dbtools import mtdb


def assimilate_proxies(prox_file):
    hlg2data = {}
    with gzip.open(prox_file, 'rt') as raw:
        for line in raw:
            if not line.startswith('#'):
                d = line.rstrip().split('\t')
                hlg, lntmd, pds = int(d[1]), float(d[3]), float(d[4])
                gcl, mmi, mmp, csb = float(d[5]), float(d[6]), float(d[7]), float(d[8])
                hlg2data[hlg] = (lntmd, gcl, mmi, mmp, csb, pds,)
    return hlg2data

def count_transcripts(gff_path):
    gff = gff2list(gff_path)
    error = False
    aliases, alia_comp = [], re.compile(gff3Comps()['Alias'])
    for line in gff:
        try:
            alias_init = line['attributes'].find('Alias=')
            alias_end = line['attributes'][alias_init:].find(';')
            if alias_end != -1:
                alias_end += alias_init
                alias = line['attributes'][alias_init + 7:alias_end]
            else:
                alias = line['attributes'][alias_init + 7:]
#            alias = alia_comp.search(line['attributes'])[1]
        except TypeError:
            if not error:
                eprint(f'WARNING: {gff_path} missing alia', flush = True)
            error = True
            continue
        if alias:
            aliases.append(alias)
    alia = set(aliases)
    return len(alia)

def calc_shannon(top_anns):
    len_anns = len(top_anns)
    ann_counts = Counter(top_anns)
    return -sum([v/len_anns * log(v/len_anns) for v in ann_counts.values()])
            

def calc_stats(ome, f, gff_path, alia = None):
    genes2hg = {}
    hlgs = []
    overall_anns = []
    if not os.path.isfile(f):
        return None, None, None, None, None, None, None, None, None
 #   ind_shannons = []
    with open(f, 'r') as raw:
        for line in raw:
            if not line.startswith('#'):
                try:
                    data = line.rstrip().split('\t')
                    clus_id, hgs, genes, t_hlgs = data[0],data[1],data[2],data[3]
                    if len(data) > 4: # there are annotations
                        raw_annotations = data[4]
                        top_anns = [x.split('|')[0] for x in raw_annotations.split(';') if x]
#                        clus_shannon = calc_clus_shannon(top_anns)
                    else:
#                        clus_shannon = None
                        top_anns = []
                except ValueError:
                    raise ValueError(line)
                    sys.exit(4)
                try:
                    hgs = [int(x) if x else None for x in hgs.split(',')]
                except ValueError:
                    eprint(f'WARNING: no HGs for: {line}')
                    hgs = []
                genes = genes.split(',')
                t_hlgs = [int(x) for x in t_hlgs.split(',')]
                for i, g in enumerate(genes):
                    genes2hg[g] = hgs[i]
                hlgs.extend(t_hlgs)
#                ind_shannons.append(clus_shannon)
                overall_anns.extend(top_anns)

    genes_in_clus = sorted(set(genes2hg.keys()))
    tot_gic = len(genes_in_clus)
    mean_gic = tot_gic/len(genes_in_clus)
    genes_in_clus.sort()
    median_gic = genes_in_clus[round(len(genes_in_clus)/2) - 1]
    if overall_anns:
        alpha = calc_shannon(overall_anns)
    else:
        alpha = None

    if not alia:
        alia = count_transcripts(gff_path)
    perc_clus = tot_gic/alia

    return ome, hlgs, mean_gic, median_gic, \
           perc_clus, tot_gic, alia, alpha, genes2hg
    

def hg2alien(genes, hg_file, ome2tax):
    if not os.path.isfile(hg_file):
        return None
    genes = set(genes)
    g2h = defaultdict(list)
    with open(hg_file, 'r') as raw:
        for line in raw:
            d = line.rstrip().split()
            try:
                q, s, i = d[0], d[1], d[-1]
            except IndexError:
                continue
            if q in genes:
                g2h[q].append((s, float(i)))

    ome2alien = defaultdict(list)
    for q, hits in g2h.items():
        g2h[q] = sorted(hits, key = lambda x: [1], reverse = True)
        self_ome = q[:q.find('_')]
        try:
            self_tax = ome2tax[self_ome]
        except KeyError: # skip omes without taxonomy, remove later
            continue
        in_tax, out_tax = None, None
        while in_tax is None or out_tax is None:
            s, i = hits[0]
            s_tax = s[:s.find('_')]
            if s_tax == self_tax:
                if in_tax is None:
                    in_tax = i
            elif s_tax:
                if out_tax is None:
                    out_tax = i
            del hits[0]
            if not hits:
                break
        if in_tax is None and out_tax is None:
            continue
        elif in_tax is None:
            in_tax = 0
            ome2alien[self_ome].append(0 - log(out_tax))
        elif out_tax is None:
            out_tax = 0
            ome2alien[self_ome].append(log(in_tax) - 0)

    return ome2alien
           
def alien_mngr(genes_in_clus, algn_dir, ome2tax, cpus = 1):
    hg2genes = defaultdict(list)
    for ome, genes2hg in genes_in_clus.items():
        for gene, hg in genes2hg.items():
            if hg is not None:
                hg2genes[hg].append(gene)

#    alien_res = []
 #   for hg, genes in tqdm(hg2genes.items(), total = len(hg2genes)):
  #      alien_res.append(hg2alien(genes, f'{algn_dir}{hg}.out', ome2tax))
    with mp.Pool(processes = cpus) as pool:
        alien_res = pool.starmap(hg2alien, tqdm(((genes, f'{algn_dir}{hg}.out', ome2tax) \
                                     for hg, genes in hg2genes.items()), 
                                     total = len(hg2genes)))

    ome_res = defaultdict(list)
    for ome2alien in alien_res:
        if ome2alien:
            for ome, aliens in ome2alien.items():
                ome_res[ome].extend(aliens)
    
    return ome_res

def summarize_stats(ome_stats, rank, hlg2data, db, tmp_path, ass_stats):
    ome2alia = {}
    tax_dict = defaultdict(lambda: {'tmd': [], 'gcl': [], 'genes': [],
                                    'mmi': [], 'mmp': [], 'gic': [],
                                    'csb': [], 'pds': [], 'sd': [],
                                    'mean_g': [], 'med_g': [], 'pc': [],
                                    'n50': [], 'bp': [], 'mask': []})
    tax2genes2hg = {}
    with open(tmp_path, 'w') as out:
        out.write('tip\ttmd\tgcl\tmmi\tmmp\tcsb\tpds\tpc' \
                + '\tclus_genes\tgenes\tdivers\ttaxon\tn50\tbp\tmask\n')
        if rank == 'ome':
            for ome, hlgs, mean_gic, median_gic, pc, gic, alia, shan, genes in ome_stats:
                if not ome:
                    continue
                tax2genes2hg[ome] = genes
                prox = [hlg2data[hlg] for hlg in set(hlgs) if hlg in hlg2data]
                n50, alen, mask = ass_stats[ome]['n50'], ass_stats[ome]['len'], ass_stats[ome]['mask']
                tax_dict[ome]['tmd'].extend([x[0] for x in prox])
                tax_dict[ome]['gcl'].extend([x[1] for x in prox])
                tax_dict[ome]['mmi'].extend([x[2] for x in prox])
                tax_dict[ome]['mmp'].extend([x[3] for x in prox])
                tax_dict[ome]['csb'].extend([x[4] for x in prox])
                tax_dict[ome]['pds'].extend([x[5] for x in prox])
                tax_dict[ome]['mean_g'].append(mean_gic)
                tax_dict[ome]['med_g'].append(median_gic)
                tax_dict[ome]['pc'].append(pc)
                tax_dict[ome]['sd'].append(shan)
                tax_dict[ome]['genes'].append(alia)
                tax_dict[ome]['gic'].append(gic)
                tax_dict[ome]['n50'].append(ass_stats[ome]['n50'])
                tax_dict[ome]['bp'].append(ass_stats[ome]['len'])
                tax_dict[ome]['mask'].append(ass_stats[ome]['mask'])
                ome2alia[ome] = alia
                try:
                    out.write(f'{ome}\t{sum(tax_dict[ome]["tmd"])/len(tax_dict[ome]["tmd"])}\t' \
                            + f'{sum(tax_dict[ome]["gcl"])/len(tax_dict[ome]["gcl"])}\t' \
                            + f'{sum(tax_dict[ome]["mmi"])/len(tax_dict[ome]["mmi"])}\t' \
                            + f'{sum(tax_dict[ome]["mmp"])/len(tax_dict[ome]["mmp"])}\t' \
                            + f'{sum(tax_dict[ome]["csb"])/len(tax_dict[ome]["csb"])}\t' \
                            + f'{sum(tax_dict[ome]["pds"])/len(tax_dict[ome]["pds"])}\t' \
                            + f'{pc}\t{gic}\t{alia}\t{shan}\t{ome}\t{n50}\t{alen}\t{mask}\n')
                except ZeroDivisionError:
                    continue
        else:
            for ome, hlgs, mean_gic, median_gic, pc, gic, alia, shan, genes in ome_stats:
                if not ome:
                    continue
                tax2genes2hg[ome] = genes
                tax = db[ome]['taxonomy'][rank]
                prox = [hlg2data[hlg] for hlg in set(hlgs) if hlg in hlg2data]                        
                tmd = [x[0] for x in prox]
                gcl = [x[1] for x in prox]
                mmi = [x[2] for x in prox]
                mmp = [x[3] for x in prox]
                csb = [x[4] for x in prox]
                pds = [x[5] for x in prox]
                n50, alen, mask = ass_stats[ome]['n50'], ass_stats[ome]['len'], ass_stats[ome]['mask']

                tax_dict[tax]['tmd'].extend(tmd)
                tax_dict[tax]['gcl'].extend(gcl)
                tax_dict[tax]['mmi'].extend(mmi)
                tax_dict[tax]['mmp'].extend(mmp)
                tax_dict[tax]['csb'].extend(csb)
                tax_dict[tax]['pds'].extend(pds)
                tax_dict[tax]['mean_g'].append(mean_gic)
                tax_dict[tax]['med_g'].append(median_gic)
                tax_dict[tax]['pc'].append(pc)
                tax_dict[tax]['sd'].append(shan)
                tax_dict[tax]['genes'].append(alia)
                tax_dict[tax]['gic'].append(gic)
                tax_dict[tax]['n50'].append(n50)
                tax_dict[tax]['bp'].append(alen)
                tax_dict[tax]['mask'].append(mask)
                ome2alia[ome] = alia
                try:
                    out.write(f'{ome}\t{sum(tmd)/len(tmd)}\t{sum(gcl)/len(gcl)}\t' \
                        + f'{sum(mmi)/len(mmi)}\t{sum(mmp)/len(mmp)}\t' \
                        + f'{sum(csb)/len(csb)}\t{sum(pds)/len(pds)}\t' \
                        + f'{pc}\t{gic}\t{alia}\t{shan}\t{tax}\t{n50}\t{alen}\t{mask}\n')
                except ZeroDivisionError:
                    continue

    return tax_dict, ome2alia, tax2genes2hg

def output_stats(out_file, tax_dict, run_alien = False):
    tax_dict = {k: v for k, v in sorted(tax_dict.items(), key = lambda x: x[0])}
    with open(out_file + '.mean.tsv', 'w') as out0, open(out_file + '.median.tsv', 'w') as out1:
        if run_alien:
            out0.write('#taxon\ttmd\tgcl\tmmi\tmmp\tcsb\tpds' \
                     + '\tgenes_per_clus\tperc_clustered\t' \
                     + 'genes_in_clus\ttot_genes\tdiversity\talienness\t' \
                     + 'n50\tbase_pairs\tmask\n')
            out1.write('#taxon\ttmd\tgcl\tmmi\tmmp\tcsb\tpds' \
                     + '\tgenes_per_clus\tperc_clustered\t' \
                     + 'genes_in_clus\ttot_genes\tdiversity\talienness\t' \
                     + 'n50\tbase_pairs\tmask\n')
        else:
            out0.write('#taxon\ttmd\tgcl\tmmi\tmmp\tcsb\tpds' \
                     + '\tgenes_per_clus\tperc_clustered\t' \
                     + 'genes_in_clus\ttot_genes\tdiversity\t' \
                     + 'n50\tbase_pairs\tmask\n')
            out1.write('#taxon\ttmd\tgcl\tmmi\tmmp\tcsb\tpds\t' \
                     + 'genes_per_clus\tperc_clustered\t' \
                     + 'genes_in_clus\ttot_genes\tdiversity\t' \
                     + 'n50\tbase_pairs\tmask\n')
        for tax, data in tax_dict.items():
            if len(data['tmd']) == 0:
                eprint(f'\tWARNING: {tax} has no results', flush = True)
                continue
            tmd = sum(data['tmd'])/len(data['tmd'])
            gcl = sum(data['gcl'])/len(data['gcl'])
            mmi = sum(data['mmi'])/len(data['mmi'])
            mmp = sum(data['mmp'])/len(data['mmp'])
            csb = sum(data['csb'])/len(data['csb'])
            pds = sum(data['pds'])/len(data['pds'])
            genes = sum(data['mean_g'])/len(data['mean_g'])
            tot_genes = sum(data['genes'])/len(data['genes'])
            tot_cg = sum(data['gic'])/len(data['gic'])
            pc = sum(data['pc'])/len(data['pc'])
            n50 = sum(data['n50'])/len(data['n50'])
            bp = sum(data['bp'])/len(data['bp'])
            try:
                mask = sum(data['mask'])/len(data['mask'])
            except TypeError:
                mask = None
            try:
                sd = sum(data['sd'])/len(data['sd'])
            except TypeError:
                sd = None
            if run_alien:
                alien = sum(data['alien'])/len(data['alien'])
                line = f'{tax}\t{tmd}\t{gcl}\t{mmi}\t{mmp}\t{csb}\t' \
                     + f'{pds}\t{genes}\t{pc}\t{tot_cg}\t{tot_genes}\t' \
                     + f'{sd}\t{alien}\t{n50}\t{bp}\t{mask}\n'
            else:
                line = f'{tax}\t{tmd}\t{gcl}\t{mmi}\t{mmp}\t{csb}\t' \
                     + f'{pds}\t{genes}\t{pc}\t{tot_cg}\t{tot_genes}\t' \
                     + f'{sd}\t{n50}\t{bp}\t{mask}\n'
            out0.write(line)
            
            data['tmd'].sort()
            data['gcl'].sort()
            data['mmi'].sort()
            data['mmp'].sort()
            data['csb'].sort()
            data['pds'].sort()
            data['med_g'].sort()
            data['pc'].sort()
            data['genes'].sort()
            data['gic'].sort()
            data['n50'].sort()
            data['bp'].sort()
            data['mask'].sort()
            prox_len = len(data['tmd'])
            tmd = data['tmd'][round(prox_len/2) - 1]
            gcl = data['gcl'][round(prox_len/2) - 1]
            mmi = data['mmi'][round(prox_len/2) - 1]
            mmp = data['mmp'][round(prox_len/2) - 1]
            csb = data['csb'][round(prox_len/2) - 1]
            pds = data['pds'][round(prox_len/2) - 1]
            genes = data['med_g'][round(len(data['med_g'])/2 - 1)]
            tot_genes = data['genes'][round(len(data['genes'])/2 - 1)]
            tot_cg = data['gic'][round(len(data['gic'])/2 - 1)]
            pc = data['pc'][round(len(data['pc'])/2 - 1)]
            sd = data['sd'][round(len(data['sd'])/2 - 1)]
            n50 = data['n50'][round(len(data['n50'])/2 - 1)]
            bp = data['bp'][round(len(data['bp'])/2 - 1)]
            data['mask'] = [x for x in data['mask'] if x] # IGNORES TRUE 0% MASKED
            try:
                mask = data['mask'][round(len(data['mask'])/2) - 1]
            except IndexError:
                mask = None

            if run_alien:
                data['alien'].sort()
                alien = data['alien'][round(len(data['alien'])/2) - 1]
                line = f'{tax}\t{tmd}\t{gcl}\t{mmi}\t{mmp}\t{csb}\t' \
                     + f'{pds}\t{genes}\t{pc}\t{tot_cg}\t{tot_genes}\t' \
                     + f'{sd}\t{alien}\t{n50}\t{bp}\t{mask}\n'
            else:
                line = f'{tax}\t{tmd}\t{gcl}\t{mmi}\t{mmp}\t{csb}' \
                     + f'\t{pds}\t{genes}\t{pc}\t{tot_cg}\t{tot_genes}\t{sd}' \
                     + f'\t{n50}\t{bp}\t{mask}\n'
            out1.write(line)


def calc_gamma_diversity(ome, ann_dir):
    """Calculate genome-wide Shannon diversity referencing the top hit,
    excludes genes with missing annotations"""
    null, ann_res = parse_hmm_res(ome, ann_dir, 0.001, 0.5, 1)
    gamma_diversity = calc_shannon(list(ann_res.values()))
    return ome, gamma_diversity


def add_alien_data(db, stats_file, stats_dict, rank, ome2alien):
    if rank != 'ome':
        tax2alien = defaultdict(list)
        for ome, aliens in ome2alien.items():
            tax = db[ome]['taxonomy'][rank]
            tax2alien[tax].extend(aliens)
    else:
        tax2alien = ome2alien

    for tax, aliens in tax2alien.items():
        stats_dict[tax]['alien'] = aliens

    with open(stats_file + '.tmp', 'w') as out:
        with open(stats_file, 'r') as raw:
            for line in raw:
                if line.startswith('#'):
                    out.write(line.rstrip() + '\talienness\n')
                else:
                    taxon = line.split()[0]
                    if taxon in tax2alien:
                        out.write(line.rstrip() \
                        + f'\t{sum(tax2alien[taxon])/len(tax2alien[taxon])}\n')
                    else:
                        out.write(line.rstrip() + '\tna\n')
    os.rename(stats_file + '.tmp', stats_file)

    return stats_dict


def main(out_dir, ome_dir, db, rank, gamma = False, ann_dir = None, 
         pfam = False, alien = False, gcf_only = False, cpus = 1):

    db = db.set_index()

    if gamma:
        if not ann_dir:
            ann_dir = out_dir + 'genome_ann/'
        if not os.path.isdir(ann_dir):
            os.mkdir(ann_dir)
        annotations = collect_files(ann_dir, 'out')
        omes = [os.path.basename(x[:-4]) for x in annotations]
        to_run = sorted(set(db.keys()).difference(set(omes)))
        if to_run:
            print(f'\nAnnotating {len(to_run)} full genomes')
            failed_omes = run_hmmsearch(pfam, to_run, ann_dir, cpus, db = db)
            passing_omes = sorted(set(db.keys()).difference(set(failed_omes)))
            
            ann_res = compile_hmm_res(ann_dir, passing_omes, cpus = cpus, 
                                      max_hits = 1)
        

    print('\nPreparing assembly statistics', flush = True)
    assembly_stats('', log_path = out_dir + 'assembly_stats.tsv', cpus = cpus, db = db)

    with open(out_dir + 'assembly_stats.tsv', 'r') as raw:
        ome2ass_stats = {}
        for line in raw:
            if not line.startswith('#'):  
                d = line.rstrip().split()
                ome, n50, mask, ass_len = d[0], d[1], d[-1], d[-5]
                ome2ass_stats[ome] = {'n50': int(n50), 'mask': float(mask),
                                      'len': int(ass_len)}



    print('\nCollecting data for omes', flush = True)
    tsvs_p = collect_files(ome_dir, 'tsv', recursive = True)
    tsvs = [x for x in tsvs_p if os.path.basename(os.path.dirname(x)) in db]
    gcfs = [x for x in tsvs if os.path.basename(x) == 'gcf.tsv']
    omes = [os.path.basename(os.path.dirname(x)) for x in gcfs]
    if rank != 'ome':
        ome2tax = {o: db[o]['taxonomy'][rank] for o in omes}
    else:
        ome2tax = {o: o for o in omes}
    primary_gcf = ome_dir + '../gcfs.tsv.gz'

    if not gcf_only:
        hlgs = [x for x in tsvs if os.path.basename(x) == 'hlg.tsv']
        omes = [os.path.basename(os.path.dirname(x)) for x in hlgs]
        primary_hlg = out_dir + '../hlgs.tsv.gz'
        if not os.path.isfile(primary_hlg):
            eprint('\nERROR: `hlgs.tsv.gz` not detected', flush = True)
            sys.exit(1)
   
        print('\nAssimilating proxy values for HLGs', flush = True)
        hlg2data = assimilate_proxies(primary_hlg)

    
        print('\nCollecting stats by ome', flush = True)
        with mp.Pool(processes = cpus) as pool:
            ome_stats = pool.starmap(calc_stats, ((ome, ome_dir + ome + '/hlg.tsv', db[ome]['gff3']) \
                                                  for ome in omes))

        


        print('\nSummarizing data by ' + rank, flush = True)
        tmp_file = out_dir + f'.cloci2summary.hlg.{rank}.tsv'
        hlg_stats, ome2alia, genes_in_clus = summarize_stats(ome_stats, rank, hlg2data, 
                                                             db, tmp_file, ome2ass_stats)
    #    if os.path.isfile(tmp_file):
     #       os.remove(tmp_file)
        if alien:
            print('\nCalculating HLG alienness', flush = True)
            ome2alien = alien_mngr(genes_in_clus, alien, ome2tax, cpus = cpus)
            hlg_stats = add_alien_data(db, tmp_file, hlg_stats, rank, ome2alien)
        out_f = out_dir + 'hlg_stats.' + rank
    
        print('\nWriting HLG stats', flush = True)
        output_stats(out_f, hlg_stats, alien)
    else:
        hlg2data = assimilate_proxies(primary_gcf)
        ome2alia = None


    
    if not os.path.isfile(primary_gcf):
        eprint('\nWARNING: `gcfs.tsv.gz` not detected; ignoring GCFs', flush = True)
 #       gcf2data = []
    else:
        print('\nAssimilating proxy values for GCFs', flush = True)
#        gcf2data = assimilate_proxies(primary_gcf)
        print('\nCollecting stats by ome', flush = True)
        with mp.Pool(processes = cpus) as pool:
            if ome2alia:
                ome_stats = pool.starmap(calc_stats, tqdm(((ome, f'{ome_dir}{ome}/gcf.tsv', 
                                                       db[ome]['gff3'], ome2alia[ome]) \
                                                      for ome in omes), total = len(omes)))
            else:
                ome_stats = pool.starmap(calc_stats, tqdm(((ome, f'{ome_dir}{ome}/gcf.tsv',
                                                       db[ome]['gff3']) for ome in omes),
                                                       total = len(omes)))
        print('\nSummarizing data', flush = True)
        tmp_file = out_dir + f'.cloci2summary.gcf.{rank}.tsv'
        gcf_stats, null, genes_in_clus = summarize_stats(ome_stats, rank, hlg2data, 
                                                         db, tmp_file, ome2ass_stats)
        if not ome2alia:
            ome2alia = null
        if alien:
            ome2alien = alien_mngr(genes_in_clus, alien, ome2tax, cpus = cpus)
            gcf_stats = add_alien_data(db, tmp_file, gcf_stats, rank, ome2alien)
  #      if os.path.isfile(tmp_file):
   #         os.remove(tmp_file)
        print('\nWriting GCF stats', flush = True)
        out_f = out_dir + 'gcf_stats.' + rank 
        output_stats(out_f, gcf_stats, alien)


def cli():
    ranks = ['ome', 'kingdom', 'phylum', 'subphylum', 'class', 'order', 'family', 'genus', 'species']
    parser = argparse.ArgumentParser(description = 'Summarize CLOCI output for taxonomic rank')
    parser.add_argument('-c', '--cloci', required = True, help = 'CLOCI|CLOCI/ome output')
    parser.add_argument('-r', '--rank', help = f'{ranks}; DEFAULT: ome')
    parser.add_argument('-d', '--mtdb', help = 'MycotoolsDB')
    parser.add_argument('-a', '--alien',
        help = 'Directory of HG alignments to quantify Alien Index')
    parser.add_argument('-g', '--gcf', action = 'store_true',
        help = 'Only run calculations for GCFs')
#    parser.add_argument('-g', '--gamma', action = 'store_true',
   #     help = 'Calculate gamma and beta diversity')
 #   parser.add_argument('-a', '--annotations', 
   #     help = '[-g] Directory of tbl-formatted Pfam annotations for gamma diversity, labeled <ome>.out')
  #  parser.add_argument('-p', '--pfam', help = '[-g] Pfam.hmm path')
    parser.add_argument('--cpus', type = int, default = 1)
    args = parser.parse_args()

    rank = args.rank.lower().rstrip().lstrip()
    if rank not in set(ranks):
        eprint('\nERROR: invalid -r', flush = True)
        sys.exit(3)

#    if args.gamma:
 #       findExecs(['hmmsearch', exit = set('hmmsearch')])

    if os.path.basename(os.path.dirname(format_path(args.cloci))) == 'ome':
        out_dir = mkOutput(format_path('./'), 'cloci2stats')
        ome_dir = format_path(args.cloci)
    else:
        out_dir = mkOutput(format_path(args.cloci), 'cloci2stats', suffix = None)
        ome_dir = format_path(args.cloci) + 'ome/'

    main(out_dir, ome_dir, mtdb(format_path(args.mtdb)), rank, 
         alien = format_path(args.alien), gcf_only = args.gcf, cpus = args.cpus)
  #       gamma = False, ann_dir = format_path(args.annotations), 
   #      pfam = format_path(args.pfam),    
    sys.exit(0)


if __name__ == '__main__':
    cli()



#! /usr/bin/env python3
"""
Co-occurrence Locus and Orthologous Cluster Identifier (CLOCI)
(C) Zachary Konkel, Jason Slot 2023
BSD 3-clause License
"""

#NEED to make similarity coefficient adjustable for domains and HLGs
#NEED to detect when a bypass failed
#NEED to uniformize percent outputs
#NEED filter output option
#NEED to implement genome-based null
#NEED call OrthoFinder option
#NEED use hg_dir from OrthoFinder in input_parsing/hgx2gcf, currently doesnt work
#NEED to make gcl_mngr split different families with the same GCF
#NEED locus-based GCF hypergeometric average
#NEED TO ADD PERCENTILE TO OUTPUT
#NEED to implement gff2svg
#NEED to delete excess when checkpoints are reached
    # implement tarring effectively and --compress
#NEED to rerun evo_conco if gcfs.pickle doesn't exist
   # update to remove gcf dir if logs are violated

import os
import re
import sys
import copy
import gzip
import shutil
import pickle
import hashlib
import argparse
import numpy as np
import multiprocessing as mp
from math import log
from tqdm import tqdm
from itertools import chain
from datetime import datetime
from collections import defaultdict
from mycotools.lib.kontools import \
    intro, outro, format_path, collect_files, \
    findExecs, eprint, tardir, write_json, \
    mkOutput, split_input
from mycotools.lib.biotools import \
    gff2list
from mycotools.lib.dbtools import mtdb, primaryDB
#from mycotools.gff2svg import main as gff2svg
from mycotools import db2microsyntree
from cloci.lib import treecalcs, evo_conco, \
     hgx2hlgs, input_parsing, hgp2hgx, generate_nulls, output_data

# if numba is available, let's import it
try:
    # this is deprecated
    from numba import njit, jit
    @njit
    def est_conds(ome_arr, cooccur_array):
        """Estimate the conditional probability a given HG combination would occur based
        on the hypergeometric distribution of each HGs' presence-absence"""
        p = 1
        for i in range(len(ome_arr) - 1): # for all but the last organisms hgx
            others, ome0 = ome_arr[i + 1:], cooccur_array[ome_arr[i]] #all others and i's hits
            other_m = cooccur_array[others, :] #grab others hits
            other_sum = np.sum(other_m, axis = 0) #sum the others' hits
            other_sum[other_sum > 0] = 1 #convert to binary
            sucs_prep = ome0 + other_sum 
            sucs = len(sucs_prep[sucs_prep > 1]) #find overlapping clusters
            tot = np.sum(ome0)
            p *= sucs/tot # conditional p = overlap/total i
        return p

except ModuleNotFoundError:
    def est_conds(ome_arr, cooccur_array):
        """Estimate the conditional probability a given HG combination would occur based
        on the hypergeometric distribution of each HGs' presence-absence"""
        p = 1
        for i in range(len(ome_arr) - 1):
            others, ome0 = ome_arr[i + 1:], cooccur_array[ome_arr[i]]
            other_m = cooccur_array[others, :]
            other_sum = np.sum(other_m, axis = 0)
            other_sum[other_sum > 0] = 1
            sucs_prep = ome0 + other_sum 
            sucs = len(sucs_prep[sucs_prep > 1])
            tot = np.sum(ome0)
            p *= sucs/tot
        return p

# deprecated
def est_combo_probs(hgx, omes, genes_in_hg_in_ome, genes_in_ome, window,
                    cooccur_arr):
    """Estimate the probability a given HGx would exist given the abundance of
    each HG and the hypergeometric probability of each HG co-occurrence"""
#    win_size = [win_size[x] for x in win_size]
#    p_coeff = calc_coeff(og0, og1, tot_genes, win_size)
    p_conds = est_conds(np.array(omes), cooccur_arr)
    p_coef = 1
    for og in hgx:
        p_coef *= est_hypergeo(omes, genes_in_hg_in_ome[og],
                               genes_in_ome, window)
    p = p_conds * p_coef
    return hgx, p

# deprecated
def est_hypergeo(
    omes, genes_in_hg_in_ome, genes_in_ome, window
    ):
    """Estimate the hypergeometric-derived probability of a particular
    HG being randomly sampled"""
    # NEED to modify to account for sliding window

    # should this be for all omes, or each og in each ome and then multiplied?
    pval = hypergeom.sf(1, genes_in_ome, genes_in_hg_in_ome, window)
    return pval

# deprecated
def combo_prob_mngr(
    hgx2omes, omes2hg2genes, omes2genes, window, cooccur_array, cpus = 1
    ):
    """Manage the estimation of the probability of randomly sampling each HG
    co-occurrence combination"""

    cmds = []
    # prepare commands for each hgx
    for hgx, omes in hgx2omes.items():
        ome = omes[-1] # coefficient ome
        genes_in_hg_in_ome = {og: len(omes2hg2genes[ome][og]) for og in hgx}
        genes_in_ome = len(omes2genes[ome])
        cmds.append([
            hgx, omes, genes_in_hg_in_ome, genes_in_ome, window,
            cooccur_array
            ])

    # run hg-by-hg, then accumulate via hgx at the end
    with mp.get_context('fork').Pool(processes = cpus) as pool: # will fail on Windows
        hypergeoRes = pool.starmap(est_combo_probs, cmds)
        pool.close()
        pool.join()

    # prepare a Bonferroni correction
    comparisons = len(hgx2omes)
    hgx2pval = {}
    for hgx, pval in hypergeoRes:
        hgx2pval[hgx] = comparisons * pval

    return hgx2pval


def threshold_hgx(hgx2omes, hgx2loc, omes2dist, 
                  ome2partition, bord_scores_list):
    """Acquire HGxs that meet minimum microsynteny tree branch length
    distributions"""
    i2hgx, hgx2i, hgx2dist, count = {}, {}, {}, 0
    for hgx, omes in hgx2omes.items():
        dist = omes2dist[omes]
        parts = set(ome2partition[x] for x in omes)
        if None in parts:
            parts = parts.remove(None)
            if not parts:
                continue
        # choose the minimum threshold [liberal]
        bord_score = min([bord_scores_list[i][len(hgx)] for i in list(parts)])
        if dist > bord_score:
             hgx2dist[hgx] = dist
             i2hgx[count], hgx2i[hgx] = hgx, count
             count += 1
             # apply the border threshold for ALL pre-grouped hgxs

    print(
        '\t\t' + str(len(i2hgx)) + ' HGx pass border threshold', 
        flush = True
        )
    hgx2omes = {v: hgx2omes[v] for v in i2hgx.values()}
    hgx2loc = {v: hgx2loc[v] for v in i2hgx.values()}
    return hgx2omes, hgx2loc, hgx2dist, i2hgx, hgx2i


def read_tune_file(tune_file, gene2hg, ome2i):
    """Read a file to tune GCF circumscription based on the genomes that should
    be included in a particular GCF"""
    with open(tune_file, 'r') as raw:
        tune_data = [x.rstrip() for x in raw if not x.startswith('#')]
    tune = {}
    for cluster in tune_data:
        try:
            name, rawgenes, rawomes, false_omes = cluster.rstrip().split('\t')
        except ValueError:
            name, rawgenes, rawomes = cluster.split('\t')
            false_omes = []
        if name in tune:
            eprint(f'\nERROR: duplicate entry for {name} in tune file',
                   flush = True)
        if ',' in rawgenes:
            genes = [x.rstrip().lstrip() for x in rawgenes.split(',')]
        else:
            genes = [x.rstrip().lstrip() for x in rawgenes.split()]
        if ',' in rawomes:
            omes = [x.rstrip().lstrip() for x in rawomes.split(',')]
        else:
            omes = [x.rstrip().lstrip() for x in rawomes.split()]
        if false_omes:
            if ',' in false_omes:
                false_omes = [x.rstrip().lstrip() for x in false_omes.split(',')]
            else:
                false_omes = [x.rstrip().lstrip() for x in false_omes.split()]
        try:
            hgs = [int(i) for i in genes]
        except ValueError:
            hgs = []
            for gene in genes:
                if gene in gene2hg:
                    hgs.append(gene)
        if not hgs:
            eprint(f'\t\tWARNING: no homology groups in {name}, skipping',
                   flush = True)
            continue
        omes = [ome2i[x] for x in omes]
        false_omes = [ome2i[x] for x in false_omes]
        if not omes:
            eprint(f'\t\tWARNING: no omes in {name}, skipping', flush = True)
        tune[name] = [tuple(sorted(set(hgs))), 
                      tuple(sorted(set(omes))),
                      tuple(sorted(set(false_omes)))]
    return tune


def main(
    db, hg_file, out_dir, plusminus = 1, hg_dir = None, hgx_dir = None,
    hgp_perc = 0.2, clus_perc = 0, hgx_perc = 0.7,
    id_perc = 30, inflation_1 = 1.1, inflation_2 = 1.3, pos_perc = 0,
    csb_thresh = 0, minimum_omes = 2, samples = 10000, pfam = None,
    min_hlg_id = 0.3, simfun = hgx2hlgs.overlap,
    constraint_path = None, aligner = 'diamond',
    run_dnds = False, cpus = 1, n50thresh = None, 
    root = None, gcl_thresh = 0, patch_thresh = 0,
    method = 'mmseqs easy-cluster', dist_thresh = 0,
    printexit = False, skipalgn = False, flag = True, partition_file = None,
    near_single_copy_genes = [], tree_path = None, 
    verbose = False, sim = 'overlap', tune_file = None,
    dist_type = 'tmd', uniq_sp = False, partition_rank = None,
    min_branch_sim= 0, algn_sens = '', min_gene_id = 30,
    fallback = False, merge_via_sim = False, ipr_path = None, force = False
    ):
    """
    The general workflow:
    log management -> input data parsing -> homology group pair (HGp) 
    identification -> microsynteny distance thresholding -> homology 
    group combination (HGx) HGx formation -> microsynteny distance 
    border thresholding -> HGx-GCF grouping ->
    microsynteny distance cluster thresholding -> PDS calculation ->
    GCL calculation -> optional dN/dS calculations (deprecated) ->
    HGx data output -> cluster retrieving -> data output
    """
    db = db.set_index()
    if uniq_sp:
        uniq_sp = db
    if not tree_path:
        tree_path = f'{out_dir}microsynt.newick'

    # determine the distance function: TMD = total microsynteny distance, MMD =
    # maximum microsynteny distance
    if dist_type == 'tmd':
        dist_func = treecalcs.calc_tmd
    else:
        dist_func = treecalcs.calc_mmd

    # import the null distribution partition data, use an established rank, or
    # generate a single null distribution for all genomes
    if partition_file:
        partition = partition_file
    elif partition_rank:
        partition = partition_rank
    else:
        partition = None

    # initialize the run by parsing and updating the log
    wrk_dir, nul_dir, inflation_1, inflation_2, gcf_ready = \
                       input_parsing.init_run(db, out_dir, 
                                              near_single_copy_genes, constraint_path,
                                              tree_path, hg_file, plusminus, 
                                              hgp_perc, #clus_perc,
                                              hgx_perc, f'{aligner}${algn_sens}', 
                                              id_perc, pos_perc, csb_thresh,
                                              patch_thresh, gcl_thresh, dist_thresh,
                                              samples, n50thresh, flag, min_gene_id,
                                              min_hlg_id, inflation_1, inflation_2, sim, 
                                              tune_file, dist_type, uniq_sp, partition,
                                              min_branch_sim, merge_via_sim, hg_dir, hgx_dir,
                                              ipr_path, pfam)

    # we are going to rerun, regardless what the log parser claims
    if force and gcf_ready:
        print('\tJust kidding', flush = True)
        gcf_ready = False

    # !! sometimes will not enter on benign changes to log
    # Skip straight to output thresholding to avoid rerunning and regenerating
    # unnecessary data structures
    if gcf_ready:
        print('\nThresholding and outputting GCFs', flush = True)
        try:
            ome_dir = out_dir + 'ome/'
            annotate = False
            if ipr_path or pfam and annotate: # check if all are annotated
                hlg_files_p = collect_files(ome_dir, 'tsv', recursive = True)
                hlg_files = [x for x in hlg_files_p \
                            if os.path.basename(x) == 'hlg.tsv']
                for hlg_f in hlg_files:
                    with open(hlg_f, 'r') as raw:
                        for line in raw:
                            if len(line.split()) < 5: # annotations missing
                                annotate = True
                            break
                    # a long filter is necessary to annotate missing
                    # annotations
                    if annotate:
                        print('\t\tAnnotations missing and requested. Long filter',
                              flush = True)
                        break
            # proceed to the long process
            if annotate:         
                ome2i, gene2hg, i2ome, hg2gene, ome2pairs, cooccur_dict = \
                    db2microsyntree.main(db, hg_file, out_dir, wrk_dir,
                                    method, tree_path, plusminus = plusminus,
                                    min_cov = 0, min_id = 0.3, n50thresh = n50thresh,
                                    near_single_copy_genes = near_single_copy_genes,
                                    constraint = constraint_path, verbose = verbose,
                                    return_post_compile = gcf_ready, cpus = cpus)
                   
                output_data.threshold_gcf_bypass(db, out_dir, wrk_dir, i2ome, gene2hg, 
                                                 dist_thresh, gcl_thresh, patch_thresh,
                                                 id_perc, pos_perc, csb_thresh, ipr_path,
                                                 pfam, cpus)
            # proceed to the short thresholding process
            else:
                output_data.threshold_gcf_quick(db, out_dir, ome_dir, dist_thresh, gcl_thresh,
                                    patch_thresh, id_perc, pos_perc, csb_thresh, cpus)
            print('\nSUCCESS!', flush = True)
            sys.exit(0)

        # there are missing files that are necessary for one of the above
        # thresholding processes, so we have to proceed through the standard
        # script
        except FileNotFoundError:
            eprint('\t\tWARNING: necessary files missing; must resume run', flush = True)
            raise FileNotFoundError
            ome2i, gene2hg, i2ome, hg2gene, ome2pairs, cooccur_dict = \
                db2microsyntree.main(db, hg_file, out_dir, wrk_dir,
                                    method, tree_path, plusminus = plusminus,
                                    min_cov = 0, min_id = 0.3, n50thresh = n50thresh,
                                    near_single_copy_genes = near_single_copy_genes,
                                    constraint = constraint_path, verbose = verbose,
                                    return_post_compile = gcf_ready, cpus = cpus)

    # build the microsynteny tree, and toss-out any genomes that
    # failed to have any overlapping HGps with others
    else:
        ome2i, gene2hg, i2ome, hg2gene, ome2pairs, cooccur_dict = \
            db2microsyntree.main(db, hg_file, out_dir, wrk_dir,
                                method, tree_path, plusminus = plusminus,
                                min_cov = 0, min_id = 0.3, n50thresh = n50thresh,
                                near_single_copy_genes = near_single_copy_genes,
                                constraint = constraint_path, verbose = verbose,
                                return_post_compile = gcf_ready, cpus = cpus)
    
    # read a tune file to tune GCF formation downstream
    if tune_file:
        print('\tReading tune file', flush = True)
        tune = read_tune_file(tune_file, gene2hg, ome2i)
    else:
        tune = None

    # compile the microsynteny tree
    print('\tReading microsynteny tree', flush = True)
    phylo = input_parsing.compile_tree(
        i2ome, tree_path, root = root
        )
    # generate null distributions for HGps
    partition_omes, ome2partition, omes2dist, min_pair_scores = \
                                generate_nulls.gen_pair_nulls(
                                            db, phylo, ome2i, wrk_dir,
                                            nul_dir, hgp_perc, ome2pairs,
                                            i2ome, samples = samples,
                                            partition_file = partition_file,
                                            partition_rank = partition_rank,
                                            uniq_sp = uniq_sp, dist_func = dist_func,
                                            cpus = cpus
                                            )
    # seed clusters by calcuating total microsynteny distance for HGps
    print('\nIII. Seeding HG pairs (HGp)', flush = True) 
    if not os.path.isfile(out_dir + 'hgps.tsv.gz'):
        print('\tCalculating seed HG-pair TMDs', flush = True)
        seed_score_start = datetime.now()
        # we are looking at all genomes
        if not uniq_sp:
            omes2dist = treecalcs.update_dists(phylo, cooccur_dict, cpus, omes2dist,
                                               func = dist_func)
        # we are only looking at genomes of unique species
        else:
            omes2dist = treecalcs.update_dists(phylo, cooccur_dict, cpus, omes2dist,
                                               func = treecalcs.calc_tmd,
                                               uniq_sp = db, i2ome = i2ome)
        # output the updated dictionary that relates a sorted tuple of ome
        # indices to their microsynteny distance (TMD/MMD) between them
        with open(wrk_dir + 'omes2dist.pickle', 'wb') as out:
            pickle.dump(omes2dist, out)

        # identify the HGps that pass thresholding
        top_hgs = []
        for hgp, omes in cooccur_dict.items():
            score = omes2dist[omes]
            parts = set(ome2partition[x] for x in omes)
            if None in parts:
                parts = parts.remove(None)
                if not parts:
                    continue
            # choose the minimum threshold [liberal]
            min_pair_score = min([min_pair_scores[i] for i in list(parts)])
            if score >= min_pair_score:
                top_hgs.append([
                    hgp[0], hgp[1], len(omes), score#, score/ome_dist
                    ])

        # write scores            
        print('\t\t' + str(datetime.now() - seed_score_start), flush = True)
        top_hgs = sorted(top_hgs, key = lambda x: x[3])
        print('\tWriting HGp TMDs', flush = True)
        with gzip.open(out_dir + 'hgps.tsv.gz', 'wt') as out:
            out.write('#hg0\thg1\tcooccurrences\tscore\tadj_score\n')
            for line in top_hgs:
                out.write('\t'.join([str(x) for x in line]) + '\n')
        print('\t\t' + str(len(top_hgs)) + ' significant HG pairs', flush = True)
        top_hgs = [(x[0], x[1]) for x in top_hgs]
    # load previous HGps
    elif not os.path.isfile(wrk_dir + 'hgx2loc.pickle'):
        print('\tLoading previous seed HGps', flush = True)
        top_hgs = input_parsing.load_seedScores(out_dir + 'hgps.tsv.gz')
        print('\t\t' + str(len(top_hgs)) + ' significant HG pairs', flush = True)
    else:
        top_hgs = None

    # begin sifting for HGxs using pairs as seeds for HGx detection
    print('\nIV. Sprouting high order HG combinations (HGx)', flush = True)
    hgx2omes, hgx2loc = hgp2hgx.hgp2hgx(db, wrk_dir, top_hgs,
                                        gene2hg, ome2i, phylo, 
                                        plusminus, cpus) 

    print('\tCalculating HGx microsynteny distances', flush = True)
    hgx_start = datetime.now()
    hgx_obs = len(hgx2omes)
    print('\t\t' + str(hgx_obs) + ' observed HGx', flush = True)

    # calculate HGx microsynteny distances
    if not uniq_sp:
        omes2dist = treecalcs.update_dists(phylo, hgx2omes, cpus, omes2dist = omes2dist,
                                           func = dist_func)
    else:
        omes2dist = treecalcs.update_dists(phylo, hgx2omes, cpus, omes2dist = omes2dist,
                                           func = treecalcs.calc_tmd, uniq_sp = db,
                                           i2ome = i2ome)
    with open(wrk_dir + 'omes2dist.pickle', 'wb') as out:
        pickle.dump(omes2dist, out)
    print('\t\t' + str(datetime.now() - hgx_start), flush = True)

    # deprecated mechanism of estimating the probability of randomly sampling
    # an HGx from the genomes provided - needs conceptual work
    calc_hgx_p = False
    if not os.path.isfile(wrk_dir + 'hgx2pval.pickle'):
        if calc_hgx_p:
            print('\tEstimating HGx probability', flush = True)
            omes2hg2genes, omes2genes = {}, defaultdict(list)
            for gene, og in gene2hg.items():
                ome = gene[:gene.find('_')]
                omeI = ome2i[ome]
                if omeI not in omes2hg2genes:
                    omes2hg2genes[omeI] = defaultdict(list)
                omes2hg2genes[omeI][og].append(gene)
                omes2genes[omeI].append(gene)
            unadjHGx2pval = combo_prob_mngr(
                hgx2omes, omes2hg2genes, omes2genes, (plusminus*2)-1, 
                cooccur_array, cpus = cpus
                )
   #     comparisons = len(hgx2omes)
  #      hgx2pval = {
 #           k: v * comparisons for k, v in unadjHGx2pval.items()
#            } # apply bonferroni correction
            hgx2pval = unadjHGx2pval
            with open(wrk_dir + 'hgx2pval.pickle', 'wb') as out:
                pickle.dump(unadjHGx2pval, out)
    else:
        with open(wrk_dir + 'hgx2pval.pickle', 'rb') as raw:
            hgx2pval = pickle.load(raw)
    

    # prepare null distributions for each size (# of OGs) observed
    # in HGxs    
    max_hgx_size = max([len(x) for x in hgx2omes])
    bord_scores_list, clus_scores_list = generate_nulls.partitions2hgx_nulls(
                                            db, partition_omes, ome2i, i2ome, gene2hg,
                                            max_hgx_size, plusminus, hgx_perc, 
                                            clus_perc, nul_dir, omes2dist, 
                                            phylo, samples, dist_func = dist_func,
                                            uniq_sp = uniq_sp, cpus = cpus)


    # collect for normalizing relative to the whole dataset later
    # should I move into absolute space? it may just be better for comparing
    # datasets in the future and some algorithm is going to pick up that gene clusters
    # are in some absolute microsynteny distance, so long as its not normalized. 

    hgx2omes, hgx2loc, hgx2dist, i2hgx, hgx2i = threshold_hgx(hgx2omes, hgx2loc,
                                                    omes2dist, ome2partition, 
                                                    bord_scores_list)
    with open(wrk_dir + 'hgx2dist.pickle', 'wb') as out:
        pickle.dump(hgx2dist, out)
    output_data.output_hgxs(hgx2dist, hgx2omes, hgx2i, i2ome, out_dir)

    hgx_dir = f'{wrk_dir}hgx/'

    # placeholder for legacy runs
    if not os.path.isfile(f'{wrk_dir}hgx2omes.pickle'):
        with open(wrk_dir + 'hgx2omes.pickle', 'wb') as out:
            pickle.dump(hgx2omes, out)

    print('\nV. Circumscribing homologous locus groups (HLGs) from HGxs', flush = True)
    if not os.path.isfile(wrk_dir + 'hlgs.pickle'): # need to add this to
    # log parsing
        # Group hgxs
        hlgs, hlg_hgxs, hlg_omes, hlg2clan, omes2dist = hgx2hlgs.classify_hlgs(
            hgx2loc, db, gene2hg, i2hgx, hgx2i, phylo, 
            ome2i, hgx2omes, hg_dir, hgx_dir, wrk_dir, ome2partition, 
            bord_scores_list, hg2gene, tune = tune, algorithm = aligner,
            omes2dist = omes2dist, clusplusminus = plusminus, 
            inflation_1 = inflation_1, inflation_2 = inflation_2,
            min_loc_id = min_hlg_id, algn_sens = algn_sens,
            min_omes = 2, cpus = cpus, simfun = simfun, printexit = printexit,
            dist_func = dist_func, uniq_sp = uniq_sp, min_branch_sim = min_branch_sim,
            skipalgn = skipalgn, minid = min_gene_id, fallback = fallback,
            merge_via_sim = merge_via_sim
            )
        with open(wrk_dir + 'hlgs.pickle', 'wb') as pickout:
            pickle.dump([hlgs, hlg_omes, hlg_hgxs, hlg2clan], pickout)
        with open(wrk_dir + 'omes2dist.pickle', 'wb') as pickout:
            pickle.dump(omes2dist, pickout)
    else: # or just load old data
        with open(wrk_dir + 'hlgs.pickle', 'rb') as raw:
            hlgs, hlg_omes, hlg_hgxs, hlg2clan = pickle.load(raw)
        # placeholder for legacy data
#        if not isinstance(hlgs, dict):
 #           hlgs = {i: v for i, v in enumerate(hlgs)}
  #          hlg_omes = {i: v for i, v in enumerate(hlg_omes)}
   #         hlg_hgxs = {i: v for i, v in enumerate(hlg_hgxs)}
    #        hlg2clan = {i: v for i, v in enumerate(hlg2clan)}

    if not uniq_sp:
        omes2dist = treecalcs.update_dists(phylo, {i: omes for i, omes in hlg_omes.items()}, 
                                        cpus, omes2dist, func = dist_func)
    else:
        omes2dist = treecalcs.update_dists(phylo, {i: omes for i, omes in hlg_omes.items()},
                                        cpus, omes2dist, func = treecalcs.calc_tmd,
                                        uniq_sp = db, i2ome = i2ome)
    with open(wrk_dir + 'omes2dist.pickle', 'wb') as out:
        pickle.dump(omes2dist, out)


#    if dist_thresh:
 #       print(f'\t{len(hlgs)} HLGs pass threshold', flush = True)
    # this is a crude pseudo cluster family expectedness check; its pseudo because the HGx
    # is now the conglomerate of the extracted loci's, and if that locus is derived from
    # multiple HGxs or the set of omes with that locus is truncated relative to the original
    # HGx then it is not obtained in the same way the null distribution was constructed;
    # nevertheless, this is a liberal mechanism that will remove edge cases to reduce computation
    # If the user doesn't want this, then remove the family percentile argument.
 #   todel = []
#    for i, omes in hlg_omes.items():
  #      dist = omes2dist[omes]
   #     parts = set(ome2partition[x] for x in omes)
    #    if None in parts:
     #       parts = parts.remove(None)
      #      if not parts:
       #         continue
        # get the threshold corresponding to the maximum sized HGx 
        # in the family, and the minimum lineages' value [liberal]
#        max_len = max([len(x) for x in hlgs[i]])
   #     max_len = len(hlg_hgxs[i]) # get the len of the complete HGx
        # if its too big, we cant evaluate its pseudo-expectedness, so pass
    #    if max_len <= plusminus * 2 + 1:
     #       clus_score = min([clus_scores_list[v][max_len] \
      #                        for v in list(parts)])
       #     if not dist > clus_score:
        #        todel.append(i)

#    todel.reverse()
 #   for i in todel:
  #      del hlg_omes[i]
   #     del hlgs[i]
    #    del hlg_hgxs[i]
     #   del hlg2clan[i]

    print('\nVI. Quantifying GCL and intra-HLG alignment similarity', flush = True)
    hgx2omes2gcl, hgx2omes2id, hgx2omes2pos, hlgs, hlg_omes = evo_conco.gcl_main( 
                            hgx2loc, wrk_dir, ome2i, hg_dir, hgx_dir,
                            aligner, db, gene2hg, plusminus, hg2gene, phylo,
                            old_path = 'mmi.pickle',
                            hlgs = hlgs, hlg_hgxs = hlg_hgxs, hlg2clan = hlg2clan,
                            minid = min_gene_id, fallback = fallback,
                            hlg_omes = hlg_omes, printexit = printexit,
                            skipalgn = skipalgn, cpus = cpus
                            )

    runOmes = [
        omes for omes in hlg_omes.values()
        ] # omes without pds scores
    runHGxs = [
        hgx for i, hgx in hlg_hgxs.items()
        ] # hgxs without pds scores

    print('\nVII. Quantifying PDS', flush = True)
    omes2patch = treecalcs.patch_main(phylo, runOmes, wrk_dir,
                            old_path = 'pds.pickle', 
                            cpus = cpus) # could make more efficient by skipping redos

    # this is the output directory for hgx2hlg and evo_conco
    hg_dir = f'{wrk_dir}hg/'

    # deprecated placeholder for dnds methodology
    hgx2dnds = {}
    if not uniq_sp:
        omes2dist = treecalcs.update_dists(phylo, 
                                        {i: v for i, v in hlg_omes.items()}, 
                                        cpus, omes2dist, func = dist_func) 
    else:
        omes2dist = treecalcs.update_dists(phylo,
                                        {i: v for i, v in hlg_omes.items()},
                                        cpus, omes2dist, func = treecalcs.calc_tmd,
                                        uniq_sp = db, i2ome = i2ome)
    with open(wrk_dir + 'omes2dist.pickle', 'wb') as out:
        pickle.dump(omes2dist, out)


    print('\nIIX. Writing and annotating clusters', flush = True)
    output_data.output_hlgs(db, wrk_dir, hlgs, hlg_omes, 
         i2ome, out_dir, hlg_hgxs,
         omes2dist, omes2patch, hgx2omes2gcl, hgx2omes2id,
         hgx2omes2pos, gene2hg, plusminus, ome2i,
         hlg2clan, dist_thresh, gcl_thresh, patch_thresh, id_perc,
         pos_perc, csb_thresh, ipr_path = ipr_path, pfam_path = pfam, dnds_dict = {}, 
         cpus = cpus)


def cli():
    # need these here because spawn mp context forces reimport
#    from scipy.stats import hypergeom
 #   from dna_features_viewer import GraphicFeature, GraphicRecord
    null_ranks = ['kingdom', 'phylum', 'subphylum', 'class', 
                  'order', 'family', 'genus', 'species']
    aligners = ['mmseqs', 'diamond', 'blastp']
    description = \
           """The Co-occurrence Locus and Orthologous Cluster Identifier
              (CLOCI) identifies groups of homologous loci. CLOCI enriches
              gene cluster families from groups of homologous loci in a 
              function-agnostic framework, which allows for de novo
              prediction of novel biosynthetic and non-biosynthetic cluster 
              classes. CLOCI infers the boundaries of clusters using
              evidence of unexpectedly shared microsynteny, thereby
              facilitating evidence-based locus boundary predictions."""
    parser = argparse.ArgumentParser(description = description)
    i_opt = parser.add_argument_group('Input parameters')
    i_opt.add_argument('-d', '--database', required = True, default = primaryDB(), 
        help = 'MycotoolsDB. DEFAULT: masterdb')
 #   parser.add_argument('-i', '--input', 
#        help = 'Precomputed whitespace delimitted file of homologous sequences')
    i_opt.add_argument('--pfam', help = 'Pfam-A.hmm for Pfam annotations')
    i_opt.add_argument('--interpro', help = 'interproscan.sh for IPR annotations')

    mt_opt = parser.add_argument_group('Microsynteny tree parameters')
    mt_opt.add_argument('-f', '--focal_genes',
        help = 'File of genes for neighborhood extraction of microsynteny ' \
             + 'tree')
    mt_opt.add_argument('-c', '--constraint',
        help = 'Constrain microsynteny topology to species tree w/ome code tips')
    mt_opt.add_argument('-r', '--root', 
        help = 'Ome(s) to root upon - MRCA of multiple; DEFAULT: midpoint')
    mt_opt.add_argument('-t', '--tree',
        help = 'Precomputed microsynteny tree path')

    det_opt = parser.add_argument_group('Detection parameters')
    det_opt.add_argument('-of', '--orthofinder',
        help = 'Precomputed OrthoFinder output directory')
    det_opt.add_argument('-i', '--input',
        help = 'Precomputed cluster results file')
    det_opt.add_argument('-l', '--linclust', action = 'store_true',
        help = 'Less sensitive Homology inference via linclust; DEFAULT: mmseqs cluster')
    det_opt.add_argument('-w', '--window', default = 2, type = int,
        help = 'Max genes +/- for each locus window. DEFAULT: 2 (5 gene window)')
    det_opt.add_argument('-mmd', '--maximum_dist', action = 'store_true',
        help = 'Calculated maximum microsynteny distance; DEFAULT: total microsynteny distance')
    det_opt.add_argument('-us', '--unique_sp', action = 'store_true',
        help = 'Only consider one genome for replicate species in microsynteny distance' \
             + ' calculations', default = False)

    nul_opt = parser.add_argument_group('Null parameters')
    nul_opt.add_argument('-nr', '--null_rank', default = 'species',
        help = f'Taxonomic rank for local null models {null_ranks}; DEFAULT: species')
    nul_opt.add_argument('-np', '--null_partitions', 
        help = 'Tab-delimited file with omes for each null sample group on ' \
             + 'separate lines "#<LINEAGE>\\n<OME1> <OME2> <OMEn>"')
    nul_opt.add_argument('-ns', '--null_sample', type = int, default = 10000,
        help = 'Samples for null distributions; DEFAULT: 10,000')


    fam_opt = parser.add_argument_group('Locus aggregation parameters')
    fam_opt.add_argument('-a', '--aligner', default = 'diamond',
        help = f'Alignment algorithm: {aligners}; DEFAULT: diamond')
    fam_opt.add_argument('-sa', '--sensitive_align',
        help = f'[diamond ultra-sensitive] or [mmseqs -s 7.5 --num_iterations 3]')
    fam_opt.add_argument('-s', '--similarity', default = 'sorensen',
        help = 'HLG similarity metric: [J]accard, [O]verlap, [S]orensen; ' \
             + 'DEFAULT: Sorensen')
    fam_opt.add_argument('-mg', '--minimum_gene_id', type = float, default = 45,
        help = 'Percent [30 < value < 100] ID minimum between gene for loci; ' \
             + 'DEFAULT: 45')
    fam_opt.add_argument('-ml', '--minimum_loc_id', type = float, default = 30,
        help = 'Percent [0 < value < 100] ID minimum between loci for HLG ' \
             + 'DEFAULT: 30')
#    fam_opt.add_argument('-tm', '--topology_merge', action = 'store_true', default = False,
 #       help = 'Merge loci with similar topologies prior to 1st MCL round via -ts')
    fam_opt.add_argument('-ts', '--min_topology_sim', type = float, default = 25,
        help = 'Percent [0 < value < 100] topology similarity (Jaccard) minimum ' \
             + 'for singleton merging AND merging loci prior to HLG aggregation; ' \
             + 'DEFAULT: 25')
#    fam_opt.add_argument('-I', '--inflation', default = 1.3, type = float,
 #       help = 'MCL inflation during family detection; DEFAULT: 1.3')
    fam_opt.add_argument('-I1', '--inflation_rnd1', default = 1.1, type = float,
        help = 'MCL inflation 1: affects domain/merging granularity; DEFAULT: 1.1')
    fam_opt.add_argument('-I2', '--inflation_rnd2', default = 1.3, type = float,
        help = 'MCL inflation 2: affects HLG/GCF granularity; DEFAULT: 1.3')
    fam_opt.add_argument('-T', '--tune', 
        help = 'Tune inflation to subset data of clusters represented ' \
             + 'in a tab-delimited ' \
             + 'file "<CLUSTER>\\t<CONSERVED_HGS/GENES>\\t<OMES>\\t<EXCLUDED_OMES>"')
   

    thr_opt = parser.add_argument_group('GCF filtering parameters')
    thr_opt.add_argument('-hp', '--hgp_percentile', type = int, default = 20,
        help = 'Null percentile of HG pair distances; DEFAULT: 20')
    thr_opt.add_argument('-xp', '--hgx_percentile', type = int,
        help = 'Null percentile [0 < value < 100] of HGx microsynteny distances. ' \
             + 'Must be less than -fp; DEFAULT: 61', default = 61)
#    thr_opt.add_argument('-fp', '--gcf_percentile', type = int, default = 0, 
 #       help = 'Pseudonull percentile [0 < value < 100] of GCF microsynteny distances')
    thr_opt.add_argument('-ip', '--id_percent', default = 0, type = float,
        help = 'Percent [0 < value < 100] identity minimum for gene cluster family')
    thr_opt.add_argument('-pp', '--pos_percent', default = 0, type = float,
        help = 'Percent [0 < value < 100] positive minimum for gene cluster family')
    thr_opt.add_argument('-ct', '--csb_threshold', default = 0, type = float,
        help = "Threshold [0 < value < 1] conservative substitution bias minimum for " \
             + ' gene cluster family')
    thr_opt.add_argument('-pt', '--pds_threshold', default = 0, type = float,
        help = "Threshold [0 < value < 1] of gene cluster family " \
             + " phylogenetic distribution sparsity")
    thr_opt.add_argument('-gt', '--gcl_threshold', default = 0, type = float,
        help = "Threshold [0 < value < 1] of gene cluster committment")
    thr_opt.add_argument('-tt', '--md_threshold', default = 0, type = float,
        help = "Threshold [0 < value < 1] of log-normalized GCF TMDs")

    run_opt = parser.add_argument_group('Runtime parameters')
#    run_opt.add_argument('-s', '--dnds', action = 'store_true', help = 'Run dN/dS calculations')
    run_opt.add_argument('--n50', help = 'Minimum assembly N50')
    run_opt.add_argument('--stop', action = 'store_true', 
                        help = 'Export HG alignment commands ' \
                             + 'for parallelization and stop')
    run_opt.add_argument('--skip', action = 'store_true', 
                        help = 'Ignore missing HG alignments as assumed failures')
    run_opt.add_argument('--fallback', action = 'store_true',
                        help = 'Fallback to diamond from failed alignments')
    run_opt.add_argument('-n', '--new', action = 'store_true', 
        help = 'Rerun with new parameters and overwrite incompatible data')
    run_opt.add_argument('--force', action = 'store_true',
        help = 'Force rerun over bypassing')
    run_opt.add_argument('--compress', action = 'store_true', 
        help = 'Compress run; SEMI-FUNCTIONAL')
    run_opt.add_argument('--cpus', default = mp.cpu_count(), type = int,
                         help = 'DEFAULT: all')

    dir_opt = parser.add_argument_group('Alternative directories')
    dir_opt.add_argument('-o', '--output_dir', 
        help = 'Output/resume directory; DEFAULT: cloci_YYYYmmdd')
    dir_opt.add_argument('-hg', '--hg_dir', help = 'HG faa dir, format <HG>.faa')
    dir_opt.add_argument('-hgx', '--hgx_dir', 
        help = 'HGx alignment DB and results dir, format <HG>.out and <HG>.dmnd/<HG>.mmseqs')
    args = parser.parse_args()

    # NEED to reinstate, currently does not function correctly
    if args.compress:
        if not args.output_dir:
            eprint('\nERROR: compression requires -o', flush = True)
        elif not os.path.isdir(args.output_dir):
            eprint('\nERROR: -o directory does not exist', flush = True)
        dirs = [format_path(args.output_dir) + 'working/hgx/']
        for d in dirs:
            tardir(d, True)
        sys.exit(4)

    # grab the proposed root for the microsynteny tree
    if args.root:
        root = split_input(args.root)
        root_txt = ','.join(root)
    else:
        root = []
        root_txt = 'midpoint'

 #   if not args.hgx_percentile:
#        args.hgx_percentile = args.gcf_percentile # set the default

    min_topology_sim = args.min_topology_sim/100

    # parse the similarity coefficient used in HLG circumscription
    if not args.similarity:
        simfun = hgx2hlgs.overlap # set default function
        sim = 'overlap'
    elif args.similarity.lower() in {'o', 'overlap', 'oc'}:
        simfun = hgx2hlgs.overlap
        sim = 'overlap'
    elif args.similarity.lower() in {'j', 'jaccard', 'jacard', 'jac'}:
        simfun = hgx2hlgs.jaccard
        sim = 'Jaccard'
    elif args.similarity.lower() in {'s', 'sorensen', 'sorenson', 'sor'}:
        simfun = hgx2hlgs.sorensen
        sim = 'Sorensen'
    else:
        eprint(f'\nERROR: invalid -s: {args.similarity}',
               flush = True)
        sys.exit(43)

    # check for dependency presence
    execs = ['mafft', 'mcl',
             'mcxdump', 'mcxload', 'iqtree']
    if args.aligner.lower() not in aligners:
        eprint(f'\nERROR: invalid -a: {args.aligner}',
               flush = True)
        sys.exit(47)
    else:
        aligner = args.aligner.lower()
        if args.sensitive_align:
            algn_sens = 'sensitive'
        else:
            algn_sens = ''
        execs.append(aligner)
        if aligner == 'mmseqs' and args.pos_percent:
            eprint(f'\nERROR: -a mmseqs and -pp are incompatible',
                   flush = True)
            sys.exit(613)

    # choose the annotation software
    # NEED to implement interpro
    if args.pfam and args.interpro:
        eprint('\nERROR: --pfam and --interpro are incompatible',
               flush = True)
        sys.exit(921)
    elif args.pfam:
        execs.append('hmmsearch')
    if args.interpro:
        execs.append(format_path(args.interpro))

    # if the proposed alignment methodology fails, implement Diamond as a
    # back-up
    # NEED to remove MMseqs as an option
    if args.fallback:
        if 'diamond' not in set(execs):
            execs.append('diamond')

    # parse OrthoFinder data for the homology groups
    if args.orthofinder:
        of_out = format_path(args.orthofinder)
        if os.path.isdir(of_out):
            homogroups = of_out + '/Orthogroups/Orthogroups.txt'
            hg_dir = of_out + '/Orthogroup_Sequences/'
        else:
            homogroups = of_out
            hg_dir = os.path.dirname(of_out) + '../Orthogroup_Sequences/'
        if not os.path.isfile(hg_dir + 'OG0000000.fa'):
            hg_dir = format_path(args.hg_dir)
        method = 'orthofinder'
    # parse an inputted set of homology groups
    elif args.input:
        homogroups = format_path(args.input)
        hg_dir = format_path(args.hg_dir)
        method = 'mmseqs easy-cluster'
    # circumscribe homology groups using linclust
    elif args.linclust:
        method = 'mmseqs easy-linclust'
        hg_dir = format_path(args.hg_dir)
        homogroups = None
        execs.append('mmseqs')
    # circumscribe homology groups using easy-cluster
    else:
        method = 'mmseqs easy-cluster'
        hg_dir = format_path(args.hg_dir)
        homogroups = None
        execs.append('mmseqs')

    hgx_dir = format_path(args.hgx_dir)

    # set the minimum ID for considering something a homolog
    if args.minimum_gene_id < 30:
        eprint('\nERROR: -mg must be > 30', flush = True)
        sys.exit(107)

    # set the tune arguments if there is a tune file
    if args.tune:
        tune_file = format_path(args.tune)
        args.inflation_rnd1 = None
        args.inflation_rnd2 = None
        if not os.path.isfile(tune_file):
            eprint('\nERROR: -T does not exist', flush = True)
            sys.exit(103)
    else:
        tune_file = None

    # set the distance type used for microsynteny distance qunatitation
    if args.maximum_dist:
        dist = 'mmd'
        if args.unique_sp:
            eprint('\nERROR: -mmd is not compatible with -us', flush = True)
            sys.exit(104)
        uniq_sp = False
    else:
        dist = 'tmd'
        uniq_sp = args.unique_sp

    # set the method of generating null distributions
    if args.null_partitions and args.null_rank:
        eprint('\nERROR: -np and -nr are incompatible', flush = True)
        sys.exit(105)
    elif args.null_partitions:
        partition_file = format_path(args.null_partitions)
        partition_rank = None
        partition = partition_file
    elif args.null_rank:
        if args.null_rank.lower() not in set(null_ranks):
            eprint(f'\nERROR: invalid rank {args.null_rank.lower()}', flush = True)
            sys.exit(106)
        partition_rank = args.null_rank.lower()
        partition = partition_rank
        partition_file = None
    else:
        partition_rank, partition_file = None, None
        partition = None

#    if not os.path.isfile(pfam):
 #       print('\nERROR: invalid Pfam-A.hmm path', flush = True)
  #      sys.exit(4)

    # check the dependencies
    findExecs(execs, exit = set(execs))
#    if args.gcf_percentile < args.hgx_percentile and args.gcf_percentile:
 #       print('\nERROR: HGx percentile is greater than GCF percentile',
  #          flush = True)
   #     sys.exit(3)
    #elif not args.gcf_percentile:
     #   args.gcf_percentile = args.hgx_percentile

    # create/check the output directory
    if not args.output_dir:
        args.output_dir = format_path('./')
        out_dir = mkOutput(format_path(args.output_dir), 'cloci')
    elif os.path.isdir(args.output_dir):
        out_dir = format_path(args.output_dir)
    else:
        os.mkdir(format_path(args.output_dir + '/'))
        out_dir = format_path(args.output_dir)

    # set the topological constraint for microsynteny tree reconstruction
    if args.constraint:
        constraint_path = format_path(args.constraint)
    else:
        constraint_path = None

    # set the percentiles for HGps and HGxs
    if len(str(args.hgp_percentile)) > 1:
        hgp_perc = float('.' + str(args.hgp_percentile))
    else:
        hgp_perc = float('.0' + str(args.hgp_percentile))
#    if len(str(args.gcf_percentile)) > 1:
 #       clus_perc = float('.' + str(args.gcf_percentile))
  #  else:
   #     clus_perc = float('.0' + str(args.gcf_percentile))   
    if len(str(args.hgx_percentile)) > 1:
        hgx_perc = float('.' + str(args.hgx_percentile))
    else:
        hgx_perc = float('.0' + str(args.hgx_percentile))   

    # prepare the introduction dictionary
    args_dict = {
        'Homology groups': homogroups, 'Sequence clusters': method, 'MycotoolsDB': args.database, 
        'Focal genes': args.focal_genes, 
        'Microsynteny tree': format_path(args.tree), 'Toplogy constraint': constraint_path,
        'Root': root_txt, 
        'Pfam DB': format_path(args.pfam), 'InterProScan': format_path(args.interpro), 
        'Window': args.window*2+1,
        'Distance type': dist, 'Unique species': uniq_sp,
        'HGp Percentile': args.hgp_percentile, #'Precluster threshold': args.clus_threshold,
        'HGx percentile': args.hgx_percentile, 'HLG aligner': f'{aligner} {algn_sens}',
        'Similarity index': sim, 'Minimum gene id': args.minimum_gene_id,
        'Minimum locus id': args.minimum_loc_id,
        'Inflation 1': args.inflation_rnd1, 'Inflation 2': args.inflation_rnd2,
        'Tune clusters': tune_file, 'Minimum CSB': args.csb_threshold,
#        'GCF percentile': args.gcf_percentile, 
        'Minimum GCF %id': args.id_percent, 'Minimum GCF %pos': args.pos_percent,
        'PDS threshold': args.pds_threshold, 'GCL threshold': args.gcl_threshold,
        'Log Normal TMD threshold': args.md_threshold, 'Null partitions': partition,
        'Null samples': args.null_sample, #'Calculate dN/dS': args.dnds, 
        'Minimum N50': args.n50, # 'Family sensitivity': args.sensitivity,
        'Processors': args.cpus, 'Output directory': args.output_dir,
        'HG directory': hg_dir, 'HGx directory': hgx_dir,
        'Overwrite': bool(args.new)
        }
    start_time = intro('CLOCI', args_dict, 
                       'Zachary Konkel, Laura Kubatko, Jason Slot')
    date = datetime.strftime(start_time, '%Y%m%d')

    # choose focal genes for microsynteny tree reconstruction
    if args.focal_genes:
        with open(format_path(args.focal_genes), 'r') as raw:
            focal_genes = [x.rstrip() for x in raw.read().split() \
                           if x.rstrip()]
    else:
        focal_genes = []

    db = mtdb(format_path(args.database))
    main(
        db, homogroups, out_dir, plusminus = args.window,
        cpus = args.cpus, hg_dir = hg_dir, inflation_1 = args.inflation_rnd1,
        inflation_2 = args.inflation_rnd2,
        hgp_perc = hgp_perc, hgx_dir = hgx_dir, #clus_thresh = args.clus_threshold,
  #      clus_perc = clus_perc, aligner = aligner,#seed_thresh = args.seed_threshold,
        aligner = aligner, ipr_path = None, csb_thresh = args.csb_threshold,
        id_perc = args.id_percent, pos_perc = args.pos_percent,
        hgx_perc = hgx_perc, pfam = format_path(args.pfam), samples = args.null_sample,
        partition_file = args.null_partitions, #        run_dnds = args.dnds, 
        n50thresh = args.n50, near_single_copy_genes = focal_genes,
        root = root, gcl_thresh = args.gcl_threshold, 
        constraint_path = constraint_path, simfun = simfun,
        patch_thresh = args.pds_threshold, method = method,
        printexit = args.stop, skipalgn = args.skip, flag = bool(not args.new),
        min_hlg_id = args.minimum_loc_id / 100, sim = sim, 
        tree_path = format_path(args.tree), tune_file = tune_file,
        dist_thresh = args.md_threshold, uniq_sp = uniq_sp,
        dist_type = dist, partition_rank = partition_rank, 
        min_branch_sim = min_topology_sim, algn_sens = algn_sens,
        min_gene_id = args.minimum_gene_id, fallback = args.fallback,
        merge_via_sim = False, force = args.force #args.topology_merge
        )

    outro(start_time)


if __name__ == '__main__':
    cli()
